#ifndef FUPERMOD_CBLAS_H_
#define FUPERMOD_CBLAS_H_

#include "config.h"
#include "fupermod/fupermod_conf.h"
#include "fupermod/fupermod_bind.h"
#include <gsl/gsl_cblas.h>

/*!
 * \defgroup fupermod_cblas Wrappers for BLAS libraries
 * Main data structures are BLAS operation workspaces, for example \ref fupermod_gemm.
 * A BLAS operation workspace encapsulates the internal state of the operation
 * and provides the execute method, using CBLAS interface.
 * \section tests Tests and examples
 * - \ref examples/cblas_test.c
 * - \ref examples/gemm_block.c
 * \{
 */
#ifdef __cplusplus
extern "C" {
#endif

#ifdef ENABLE_BLAS_SP
  /*! Single-precision floating-point data type */
  typedef float fupermod_float;
  /*! Single-precision floating-point data type for MPI */
#define FUPERMOD_MPI_FLOAT MPI_FLOAT
#else
  /*! Double-precision floating-point data type */
  typedef double fupermod_float;
  /*! Double-precision floating-point data type for MPI */
#define FUPERMOD_MPI_FLOAT MPI_DOUBLE
#endif

  /*! GEMM workspace */
  typedef struct fupermod_gemm {
    /*! configuration */
	  fupermod_process_conf* conf;
    /* time recording*/
    float time;
    /*! params */
    void* params;
  } fupermod_gemm;

  /*! Allocate GEMM workspace */
  fupermod_gemm* fupermod_gemm_alloc(fupermod_process_conf* conf);

  /*! Execute GEMM */
  void fupermod_gemm_execute(struct fupermod_gemm* gemm,
      const enum CBLAS_ORDER Order, const enum CBLAS_TRANSPOSE TransA, const enum CBLAS_TRANSPOSE TransB,
      const int M, const int N, const int K,
      const fupermod_float alpha, const fupermod_float *A, const int lda,
      const fupermod_float *B, const int ldb,  const fupermod_float beta,
      fupermod_float *C, const int ldc);

  /*! Free GEMM workspace */
  void fupermod_gemm_free(fupermod_gemm* gemm);

#ifdef __cplusplus
}
#endif
/*!
 * \}
 */

#endif /* FUPERMOD_CBLAS_H_ */
