/*
 * This is the CPU implementation of \ref fupermod_memory interface.
 * The GPU implementation is in cuda_memory.cpp
 */
#include "config.h"
#include "fupermod_memory.h"
#ifdef HAVE_LIBCUBLAS
#include <cuda_runtime.h>
#include "cutil.h"
#endif
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

extern "C"
void* fupermod_malloc(size_t size, fupermod_process_conf* conf) {
  void* ptr = NULL;
  if (!strcmp(conf->device_type, "cpu")) {
    ptr = malloc(size);
  } else {
    if (!strcmp(conf->device_type, "gpu")) {
      #ifdef HAVE_LIBCUBLAS
      char* subopts = strdup(conf->subopts);
      char* subopts_0 = subopts;
      char* tokens[] = {"id", "md", NULL};
      char* value;
      int device_id = 0;
      int memory_mode = 0x00;
      while (*subopts != '\0') {
        switch (getsubopt(&subopts, tokens, &value)) {
          case 0:
            device_id = atoi(value);
            break;
          case 1:
            memory_mode = atoi(value);
            break;
        }
      }
      free(subopts_0);
      if (memory_mode == -1) {
        ptr = malloc(size);
      } else {
        CUDA_SAFE_CALL(cudaSetDevice(device_id));
        CUDA_SAFE_CALL(cudaHostAlloc(&ptr, size, memory_mode));
      }
      #else
      fprintf(stderr, "%s: %s: Error no cuda library\n", __FILE__, __func__);
      #endif
    }
  }
  return ptr;
}

extern "C"
void fupermod_free(void* ptr, fupermod_process_conf* conf) {
  if (!strcmp(conf->device_type, "cpu")) {
    free(ptr);
  } else {
    if (!strcmp(conf->device_type, "gpu")) {
      #ifdef HAVE_LIBCUBLAS
      // Note: cudaDeviceReset shouldn't be performed prior to get all pinned-memory freed
      // NOte: cudaDeviceReset is called in fupermod_gemm_free, so it should be invoked at the last
      char* subopts = strdup(conf->subopts);
      char* subopts_0 = subopts;
      char* tokens[] = {"id", "md", NULL};
      char* value;
      int device_id = 0;
      int memory_mode = 0x00;
      while (*subopts != '\0') {
        switch (getsubopt(&subopts, tokens, &value)) {
          case 0:
            device_id = atoi(value);
            break;
          case 1:
            memory_mode = atoi(value);
            break;            
        }
      }
      free(subopts_0);
      if (memory_mode == -1) {
        free(ptr); ptr = NULL;
      } else {
        CUDA_SAFE_CALL(cudaSetDevice(device_id));
        CUDA_SAFE_CALL(cudaFreeHost(ptr)); ptr = NULL;
      }
      #else
      fprintf(stderr, "%s: %s: Error no cuda library\n", __FILE__, __func__);
      #endif
    }
  }
}
