#include "config.h"
#include "fupermod_cblas.h"
#include <stdio.h>
#include <string.h>

#ifdef HAVE_LIBCUBLAS
#include "cuda_auxiliary.hpp"
#include "cutil.h"
#include "cutil_inline_runtime.h"

// result not correct, just want to see the difference between using overlapping or not, 1D version
void gemm_kernel_gpu_server_exceed_nd2_3(fupermod_gemm* gemm,
          const enum CBLAS_ORDER Order,
          const enum CBLAS_TRANSPOSE TransA,
          const enum CBLAS_TRANSPOSE TransB,
          const int M, const int N, const int K,
          const fupermod_float alpha,
          const fupermod_float *A, const int lda,
          const fupermod_float *B, const int ldb,
          const fupermod_float beta,
          fupermod_float *C, const int ldc)
{
  cublasxgemm_params* params = (cublasxgemm_params*)gemm->params;

  // pointer to host workspace
  fupermod_float *hA = params->workspace_WA;
  fupermod_float *hB = params->workspace_WB;

  // if pivot column is local, copy the pivot column to host workspace WA that is allocated as pinned memory
  if (A != hA) {
    int count = sizeof(fupermod_float) * K;
    int i;
    for (i = 0; i < M; i++) {
      memcpy(hA + i * K, A + i * lda, count);
    }
  }
  // if pivot row is local, copy the pivot row to host workspace WB that is allocated as pinned memory
  if (B != hB) {
      int count = sizeof(fupermod_float) * N;
      int i;
      for (i = 0; i < K; i++) {
        memcpy(hB + i * N, B + i * ldb, count);
      }
    } 
/*  if (B != hB) {
    int count = sizeof(fupermod_float) * N;
    int i;
    for (i = 0; i < K; i++) {
      memcpy(params->dma_WB + i * N, B + i * ldb, count);
    }
    hB = params->dma_WB;
  } else {
    if (params->dma_WB) { // when build fpms, dam_WB is NULL
      memcpy(params->dma_WB, params->workspace_WB, sizeof(fupermod_float) * K * N);
      hB = params->dma_WB;
    }
  }
*/  

  
  // split matrix
  int h_, w_, *height, *width, nd = params->num_device_use;
  size_t mem_size_device = gama * getDeviceMemory(params->device_id)/sizeof(fupermod_float);
  matrix_split_server_exceed(M, N, K, mem_size_device, nd, &h_, &w_, &height, &width);
  
  // initialization
  if (params->idx_outer == 0) {
    printf("Kernel: gemm_kernel_gpu_server_exceed_1D\n");
    fprintf(stderr, "h_ %d w_ %d\n", h_, w_);
    params->dA = (fupermod_float**)malloc(sizeof(fupermod_float*) * nd);
    params->dB = (fupermod_float**)malloc(sizeof(fupermod_float*) * nd);
    params->dC = (fupermod_float**)malloc(sizeof(fupermod_float*) * nd);
    
    int i, device_id = params->device_id;
    for (i = 0; i < nd; i++) {
      // set device
      CUDA_SAFE_CALL(cudaSetDevice(device_id++));
      // alloc buffers
      CUDA_SAFE_CALL(cudaMalloc((void**)(&params->dA[i]), sizeof(fupermod_float) * height[0] * K * K * 2));
      CUDA_SAFE_CALL(cudaMalloc((void**)(&params->dB[i]), sizeof(fupermod_float) * K * width[0] * K));
      CUDA_SAFE_CALL(cudaMalloc((void**)(&params->dC[i]), sizeof(fupermod_float) * height[0] * K * width[0] * K * 2));
//    CUDA_SAFE_CALL(cudaMemset((void*)params->dC[i], 0,  sizeof(fupermod_float) * height[0] * K * width[0] * K * 2));
    }
  }

  //streams and events
  cudaStream_t stream[nd][h_ * w_ / nd];
  cudaEvent_t event_a[nd][h_ * w_ / nd];
  cudaEvent_t event_b[nd][w_ / nd];
  cudaEvent_t event_c[nd][h_ * w_ / nd];
  cudaEvent_t event_send[nd][h_ * w_ / nd];
  cudaEvent_t event_recv[nd][h_ * w_ / nd];
  
  int i, j, k, device_id = params->device_id;
  for (k = 0; k < nd; k++) {
    // set device
    CUDA_SAFE_CALL(cudaSetDevice(device_id++));
    // create streams and events
    for (j = 0; j < w_ / nd; j++) {
      CUDA_SAFE_CALL(cudaEventCreateWithFlags(&event_b[k][j], cudaEventDisableTiming));
      for (i = 0; i < h_; i++) {
        if (params->is_overlap == 1) {
          CUDA_SAFE_CALL(cudaStreamCreate(&stream[k][i + j * h_]));
        } else {// without overlapping
          stream[k][i + j * h_] = 0;
        }
        CUDA_SAFE_CALL(cudaEventCreateWithFlags(&event_a[k][i + j * h_], cudaEventDisableTiming));
        CUDA_SAFE_CALL(cudaEventCreateWithFlags(&event_c[k][i + j * h_], cudaEventDisableTiming));
        CUDA_SAFE_CALL(cudaEventCreateWithFlags(&event_send[k][i + j * h_], cudaEventDisableTiming));
        CUDA_SAFE_CALL(cudaEventCreateWithFlags(&event_recv[k][i + j * h_], cudaEventDisableTiming));
      }
    }
  }
  
  int ldda = K;
  fupermod_float** dA = params->dA;
  fupermod_float** dB = params->dB;
  fupermod_float** dC = params->dC;
  cublasHandle_t* handles = params->handles;
  
  // update in an normal order, column by column
  if (params->idx_outer % 2 == 0) {
    for (j = 0; j <= w_ - nd; j += nd) {
      for (i = 0; i <= h_ - 1; i++) {
        int idx = i + j/nd * h_;
        int buf = idx % 2;
        int buf_ = (idx + 1) % 2;
        int i_ = (i + 1) % h_;
        int j_ = (i < h_ - 1) ? j : (j + nd);
        
        int id;
        // initialise
        if (idx == 0) {
          device_id = params->device_id;
          for (id = 0; id < nd; id++) {
            CUDA_SAFE_CALL(cudaSetDevice(device_id++));
            if (id > 0) {
              CUDA_SAFE_CALL(cudaStreamWaitEvent(stream[id][idx], event_send[id-1][idx], 0));
            }
            CUDA_SAFE_CALL(cudaMemcpyAsync(dB[id], addr1D(hB,i,j+id,height,width,h_,w_,K,2), sizeof(fupermod_float) * width[j+id] * K * K, cudaMemcpyHostToDevice, stream[id][idx]));
            CUDA_SAFE_CALL(cudaMemcpyAsync(addr1D(dA[id],buf,j,height,width,h_,w_,K,4), addr1D(hA,i,j,height,width,h_,w_,K,1), 
                                           sizeof(fupermod_float) * K * height[i] * K, cudaMemcpyHostToDevice, stream[id][idx]));
            if (params->idx_outer == 0) {
                CUDA_SAFE_CALL(cudaMemcpyAsync(addr1D(dC[id],buf,j,height,width,h_,w_,K,5), addr1D(C,i,j+id,height,width,h_,w_,K,3), 
                                               sizeof(fupermod_float) * width[j+id] * K * height[i] * K, cudaMemcpyHostToDevice, stream[id][idx]));
            }
            CUDA_SAFE_CALL(cudaEventRecord(event_send[id][idx], stream[id][idx]));
          }
        }
        // gemm, then send next block of A and C, and recv block of C
        if (idx < h_ * w_ / nd - 1) {
          // gemm
          device_id = params->device_id;
          for (id = 0; id < nd; id++) {
            CUDA_SAFE_CALL(cudaSetDevice(device_id++));
            cublasSetStream(handles[id], stream[id][idx]);
            //fprintf(stderr, "i %d j %d id %d\n", i, j, id);
            CUBLAS_SAFE_CALL(CUBLAS_GEMM(handles[id], trans_cblas2cublas(TransB), trans_cblas2cublas(TransA), width[j+id] * K, height[i] * K, K, &alpha, dB[id],
                          width[j+id] * K, addr(dA[id],buf,j,height,width,N,K,4), ldda, &beta, addr(dC[id],buf,j,height,width,N,K,5), width[j+id] * K));
            CUDA_SAFE_CALL(cudaEventRecord(event_a[id][idx], stream[id][idx]));
            if (i == h_ - 1) {
              CUDA_SAFE_CALL(cudaEventRecord(event_b[id][j/nd], stream[id][idx]));
            }
          }
          // send next block of B (possibly), A and C
          device_id = params->device_id;
          for (id = 0; id < nd; id++) {
            CUDA_SAFE_CALL(cudaSetDevice(device_id++));
            // wait until blocks B, A, C for the previous device have been sent
            if (id == 0) {
              CUDA_SAFE_CALL(cudaStreamWaitEvent(stream[id][idx+1], event_send[nd-1][idx], 0));
            } else {
              CUDA_SAFE_CALL(cudaStreamWaitEvent(stream[id][idx+1], event_send[id-1][idx+1], 0));
            }
            // if possible, send next block 0f B
            if (i == h_ - 1) {
              if (j < w_ - nd) {
                CUDA_SAFE_CALL(cudaStreamWaitEvent(stream[id][idx+1], event_b[id][j/nd], 0));
              }
              CUDA_SAFE_CALL(cudaMemcpyAsync(dB[id], addr1D(hB,i,j_+id,height,width,h_,w_,K,2), sizeof(fupermod_float) * width[j_+id] * K * K, cudaMemcpyHostToDevice, stream[id][idx+1]));
            }
            // send next block of A
            if (idx > 0) {
              CUDA_SAFE_CALL(cudaStreamWaitEvent(stream[id][idx+1], event_a[id][idx-1], 0));
            }
            CUDA_SAFE_CALL(cudaMemcpyAsync(addr1D(dA[id],buf_,j,height,width,h_,w_,K,4), addr1D(hA,i_,j,height,width,h_,w_,K,1), 
                                           sizeof(fupermod_float) * K * height[i_] * K, cudaMemcpyHostToDevice, stream[id][idx+1]));
            // send next block of C
            if (params->idx_outer == 0) {
              if (idx > 0) {
                CUDA_SAFE_CALL(cudaStreamWaitEvent(stream[id][idx+1], event_c[id][idx-1], 0));
              }
              CUDA_SAFE_CALL(cudaMemcpyAsync(addr1D(dC[id],buf_,j,height,width,h_,w_,K,5), addr1D(C,i_,j_+id,height,width,h_,w_,K,3), 
                                             sizeof(fupermod_float) * width[j_+id] * K * height[i_] * K, cudaMemcpyHostToDevice, stream[id][idx+1]));
            } else {
              if (idx > 0) {
                CUDA_SAFE_CALL(cudaStreamWaitEvent(stream[id][idx+1], event_c[id][idx-1], 0));
                CUDA_SAFE_CALL(cudaMemcpyAsync(addr1D(dC[id],buf_,j,height,width,h_,w_,K,5), addr1D(C,i_,j_+id,height,width,h_,w_,K,3), 
                                               sizeof(fupermod_float) * width[j_+id] * K * height[i_] * K, cudaMemcpyHostToDevice, stream[id][idx+1]));
              }
            }
            // record the sending of blocks of this device
            CUDA_SAFE_CALL(cudaEventRecord(event_send[id][idx+1], stream[id][idx+1]));
          }
          // recv block of C
          device_id = params->device_id;
          for (id = 0; id < nd; id++) {
            CUDA_SAFE_CALL(cudaSetDevice(device_id++));
            if (id == 0) {
              if (idx > 0) { 
                CUDA_SAFE_CALL(cudaStreamWaitEvent(stream[id][idx], event_recv[nd-1][idx-1], 0));
              }
            } else {
              CUDA_SAFE_CALL(cudaStreamWaitEvent(stream[id][idx], event_recv[id-1][idx], 0));
            }
            if (params->idx_outer == (params->num_iteration - 1)) { // the last iteration
              CUDA_SAFE_CALL(cudaMemcpyAsync(addr1D(C,i,j+id,height,width,h_,w_,K,3), addr1D(dC[id],buf,j,height,width,h_,w_,K,5),
                                             sizeof(fupermod_float) * width[j+id] * K * height[i] * K, cudaMemcpyDeviceToHost, stream[id][idx]));
              CUDA_SAFE_CALL(cudaEventRecord(event_c[id][idx], stream[id][idx]));
            } else { // the last two blocks of $C$ not copied back
              if (idx < h_ * w_ / nd - 2) {
                CUDA_SAFE_CALL(cudaMemcpyAsync(addr1D(C,i,j+id,height,width,h_,w_,K,3), addr1D(dC[id],buf,j,height,width,h_,w_,K,5),
                                               sizeof(fupermod_float) * width[j+id] * K * height[i] * K, cudaMemcpyDeviceToHost, stream[id][idx]));
                CUDA_SAFE_CALL(cudaEventRecord(event_c[id][idx], stream[id][idx]));
              }
            }
            CUDA_SAFE_CALL(cudaEventRecord(event_recv[id][idx], stream[id][idx]));
          }
        } else { // tail work
          // gemm
          device_id = params->device_id;
          for (id = 0; id < nd; id++) {
            CUDA_SAFE_CALL(cudaSetDevice(device_id++));
            cublasSetStream(handles[id], stream[id][idx]);
            //fprintf(stderr, "i %d j %d id %d\n", i, j, id);
            CUBLAS_SAFE_CALL(CUBLAS_GEMM(handles[id], trans_cblas2cublas(TransB), trans_cblas2cublas(TransA), width[j+id] * K, height[i] * K, K, &alpha, dB[id],
                          width[j+id] * K, addr(dA[id],buf,j,height,width,N,K,4), ldda, &beta, addr(dC[id],buf,j,height,width,N,K,5), width[j+id] * K));
          }
          // recv C
          device_id = params->device_id;
          for (id = 0; id < nd; id++) {
            CUDA_SAFE_CALL(cudaSetDevice(device_id++));
            if (id == 0) {
              if (idx > 0) {
                CUDA_SAFE_CALL(cudaStreamWaitEvent(stream[id][idx], event_recv[nd-1][idx-1], 0));
              }
            } else {
              CUDA_SAFE_CALL(cudaStreamWaitEvent(stream[id][idx], event_recv[id-1][idx], 0));
            }
            if (params->idx_outer == (params->num_iteration - 1)) { // the last iteration
              CUDA_SAFE_CALL(cudaMemcpyAsync(addr1D(C,i,j+id,height,width,h_,w_,K,3), addr1D(dC[id],buf,j,height,width,h_,w_,K,5),
                                             sizeof(fupermod_float) * width[j+id] * K * height[i] * K, cudaMemcpyDeviceToHost, stream[id][idx]));
              CUDA_SAFE_CALL(cudaEventRecord(event_recv[id][idx], stream[id][idx]));
            }
          }
        }
      }
    }
  } else { // updating in a reverse order
    for (j = w_ - nd; j >= 0; j -= nd) {
      for (i = h_ - 1; i >= 0; i--) {
        int idx = i + j/nd * h_;
        int buf = idx % 2;
        int buf_ = (idx - 1) % 2;
        int i_ = (i + h_ - 1) % h_;
        int j_ = (i > 0) ? j : (j - nd);
        
        int id;
        // initialise
        if (idx == h_ * w_ / nd - 1) {
          device_id = params->device_id;
          for (id = 0; id < nd; id++) {
            CUDA_SAFE_CALL(cudaSetDevice(device_id++));
            if (id > 0) {
              CUDA_SAFE_CALL(cudaStreamWaitEvent(stream[id][idx], event_send[id-1][idx], 0));
            }
            CUDA_SAFE_CALL(cudaMemcpyAsync(dB[id], addr1D(hB,i,j+id,height,width,h_,w_,K,2), sizeof(fupermod_float) * width[j+id] * K * K, cudaMemcpyHostToDevice, stream[id][idx]));
            CUDA_SAFE_CALL(cudaMemcpyAsync(addr1D(dA[id],buf,j,height,width,h_,w_,K,4), addr1D(hA,i,j,height,width,h_,w_,K,1), 
                                           sizeof(fupermod_float) * K * height[i] * K, cudaMemcpyHostToDevice, stream[id][idx]));
            CUDA_SAFE_CALL(cudaEventRecord(event_send[id][idx], stream[id][idx]));
          }
        }
        // gemm, then send next block of A and C, and recv block of C
        if (idx > 0) {
          // gemm
          device_id = params->device_id;
          for (id = 0; id < nd; id++) {
            CUDA_SAFE_CALL(cudaSetDevice(device_id++));
            cublasSetStream(handles[id], stream[id][idx]);
            //fprintf(stderr, "i %d j %d id %d\n", i, j, id);
            CUBLAS_SAFE_CALL(CUBLAS_GEMM(handles[id], trans_cblas2cublas(TransB), trans_cblas2cublas(TransA), width[j+id] * K, height[i] * K, K, &alpha, dB[id],
                          width[j+id] * K, addr(dA[id],buf,j,height,width,N,K,4), ldda, &beta, addr(dC[id],buf,j,height,width,N,K,5), width[j+id] * K));
            CUDA_SAFE_CALL(cudaEventRecord(event_a[id][idx], stream[id][idx]));
            if (i == 0) {
              CUDA_SAFE_CALL(cudaEventRecord(event_b[id][idx], stream[id][j/nd]));
            }
          }
          // send next block of B (possibly), A and C
          device_id = params->device_id;
          for (id = 0; id < nd; id++) {
            CUDA_SAFE_CALL(cudaSetDevice(device_id++));
            // wait until blocks B, A, C for the previous device have been sent
            if (id == 0) {
              CUDA_SAFE_CALL(cudaStreamWaitEvent(stream[id][idx-1], event_send[nd-1][idx], 0));
            } else {
              CUDA_SAFE_CALL(cudaStreamWaitEvent(stream[id][idx-1], event_send[id-1][idx-1], 0));
            }
            // if possible, send next block of B
            if (i == 0) {
              if (j > 0) {
                CUDA_SAFE_CALL(cudaStreamWaitEvent(stream[id][idx-1], event_b[id][j/nd], 0));
              }
              CUDA_SAFE_CALL(cudaMemcpyAsync(dB[id], addr1D(hB,i,j_+id,height,width,h_,w_,K,2), sizeof(fupermod_float) * width[j_+id] * K * K, cudaMemcpyHostToDevice, stream[id][idx-1]));
            }
            // send next block of A
            if (idx < h_ * w_ / nd - 1) {
              CUDA_SAFE_CALL(cudaStreamWaitEvent(stream[id][idx-1], event_a[id][idx+1], 0));
            }
            CUDA_SAFE_CALL(cudaMemcpyAsync(addr1D(dA[id],buf_,j,height,width,h_,w_,K,4), addr1D(hA,i_,j,height,width,h_,w_,K,1), 
                                           sizeof(fupermod_float) * K * height[i_] * K, cudaMemcpyHostToDevice, stream[id][idx-1]));
            // send next block of C
            if (idx < h_ * w_ / nd - 1) {
              CUDA_SAFE_CALL(cudaStreamWaitEvent(stream[id][idx-1], event_c[id][idx+1], 0));
              CUDA_SAFE_CALL(cudaMemcpyAsync(addr1D(dC[id],buf_,j,height,width,h_,w_,K,5), addr1D(C,i_,j_+id,height,width,h_,w_,K,3), 
                                             sizeof(fupermod_float) * width[j_+id] * K * height[i_] * K, cudaMemcpyHostToDevice, stream[id][idx-1]));
            }
            // record send
            CUDA_SAFE_CALL(cudaEventRecord(event_send[id][idx-1], stream[id][idx-1]));
          }
          // recv block of C
          device_id = params->device_id;
          for (id = 0; id < nd; id++) {
            CUDA_SAFE_CALL(cudaSetDevice(device_id++));
            if (id == 0) {
              if (idx < h_ * w_ / nd - 1) {
                CUDA_SAFE_CALL(cudaStreamWaitEvent(stream[id][idx], event_recv[nd-1][idx+1], 0));
              }              
            } else {
              CUDA_SAFE_CALL(cudaStreamWaitEvent(stream[id][idx], event_recv[id-1][idx], 0));
            }
            // the last iteration
            if (params->idx_outer == (params->num_iteration - 1)) {
              CUDA_SAFE_CALL(cudaMemcpyAsync(addr1D(C,i,j+id,height,width,h_,w_,K,3), addr1D(dC[id],buf,j,height,width,h_,w_,K,5),
                                             sizeof(fupermod_float) * width[j+id] * K * height[i] * K, cudaMemcpyDeviceToHost, stream[id][idx]));
              CUDA_SAFE_CALL(cudaEventRecord(event_c[id][idx], stream[id][idx]));
            } else { // the last two blocks of $C$ not copied back
              if (idx > 1) {
                CUDA_SAFE_CALL(cudaMemcpyAsync(addr1D(C,i,j+id,height,width,h_,w_,K,3), addr1D(dC[id],buf,j,height,width,h_,w_,K,5),
                                               sizeof(fupermod_float) * width[j+id] * K * height[i] * K, cudaMemcpyDeviceToHost, stream[id][idx]));
                CUDA_SAFE_CALL(cudaEventRecord(event_c[id][idx], stream[id][idx]));
              }
            }
            CUDA_SAFE_CALL(cudaEventRecord(event_recv[id][idx], stream[id][idx]));
          }
        } else { // tail work
          // gemm
          device_id = params->device_id;
          for (id = 0; id < nd; id++) {
            CUDA_SAFE_CALL(cudaSetDevice(device_id++));
            cublasSetStream(handles[id], stream[id][idx]);
            //fprintf(stderr, "i %d j %d id %d\n", i, j, id);
            CUBLAS_SAFE_CALL(CUBLAS_GEMM(handles[id], trans_cblas2cublas(TransB), trans_cblas2cublas(TransA), width[j+id] * K, height[i] * K, K, &alpha, dB[id],
                          width[j+id] * K, addr(dA[id],buf,j,height,width,N,K,4), ldda, &beta, addr(dC[id],buf,j,height,width,N,K,5), width[j+id] * K));
          }
          // recv block of C
          device_id = params->device_id;
          for (id = 0; id < nd; id++) {
            CUDA_SAFE_CALL(cudaSetDevice(device_id++));
            if (id == 0) {
              if (idx < h_ * w_ / nd - 1) {
                CUDA_SAFE_CALL(cudaStreamWaitEvent(stream[id][idx], event_recv[nd-1][idx+1], 0));
              }
            } else {
              CUDA_SAFE_CALL(cudaStreamWaitEvent(stream[id][idx], event_recv[id-1][idx], 0));
            }
            if (params->idx_outer == (params->num_iteration - 1)) { // the last iteration
              CUDA_SAFE_CALL(cudaMemcpyAsync(addr1D(C,i,j+id,height,width,h_,w_,K,3), addr1D(dC[id],buf,j,height,width,h_,w_,K,5),
                                             sizeof(fupermod_float) * width[j+id] * K * height[i] * K, cudaMemcpyDeviceToHost, stream[id][idx]));
              CUDA_SAFE_CALL(cudaEventRecord(event_recv[id][idx], stream[id][idx]));
            }
          }
        }
      }
    }
  }
  CUDA_SAFE_CALL(cudaDeviceSynchronize());

  printf("idx_outer %d\n", params->idx_outer);
  params->idx_outer++;
    
  // destroy events and streams
  device_id = params->device_id;
  for (k = 0; k < nd; k++) {
    // set device
    CUDA_SAFE_CALL(cudaSetDevice(device_id++));
    // create streams and events
    for (j = 0; j < w_ / nd; j++) {
      CUDA_SAFE_CALL(cudaEventDestroy(event_b[k][j]));
      for (i = 0; i < h_; i++) {
        if (params->is_overlap == 1) {
          CUDA_SAFE_CALL(cudaStreamDestroy(stream[k][i + j * h_]));
        } else {
          stream[k][i + j * h_] = 0;
        }
        CUDA_SAFE_CALL(cudaEventDestroy(event_a[k][i + j * h_]));
        CUDA_SAFE_CALL(cudaEventDestroy(event_c[k][i + j * h_]));
        CUDA_SAFE_CALL(cudaEventDestroy(event_send[k][i + j * h_]));
        CUDA_SAFE_CALL(cudaEventDestroy(event_recv[k][i + j * h_]));
      }
    }
  }

  // free memory
  free(height); height = NULL; free(width); width = NULL;
  // at the last iteration, free device memory and destroy handles
  if ((params->num_iteration) && (params->idx_outer >= params->num_iteration)) {
    printf("idx_outer %d, params->num_iteration %d, free device memory dA/B/C ooc server\n", params->idx_outer, params->num_iteration);
    device_id = params->device_id;
    for (j = 0; j < nd; j++) {
      CUDA_SAFE_CALL(cudaSetDevice(device_id++));
      CUDA_SAFE_CALL(cudaFree(params->dA[j])); params->dA[j] = NULL;
      CUDA_SAFE_CALL(cudaFree(params->dB[j])); params->dB[j] = NULL;
      CUDA_SAFE_CALL(cudaFree(params->dC[j])); params->dC[j] = NULL;
      CUBLAS_SAFE_CALL(cublasDestroy(params->handles[j])); params->handles[j] = NULL;
    }
    free(params->dA); params->dA = NULL;
    free(params->dB); params->dB = NULL;
    free(params->dC); params->dC = NULL;
    free(params->handles); params->handles = NULL;
  }
}
#endif
