#include "config.h"
#include "fupermod_cblas.h"
#include <stdio.h>
#include <string.h>

#ifdef HAVE_LIBCUBLAS
#include "cuda_auxiliary.hpp"
#include "cutil.h"
#include "cutil_inline_runtime.h"

// result not correct, just on device with one copy engine, 2Dasync not working
// image matrix C, and matrix wB have been re-arranged so that blocks are arranged according to idx in memory
void gemm_kernel_gpu_single_exceed_3(fupermod_gemm* gemm,
          const enum CBLAS_ORDER Order,
          const enum CBLAS_TRANSPOSE TransA,
          const enum CBLAS_TRANSPOSE TransB,
          const int M, const int N, const int K,
          const fupermod_float alpha,
          const fupermod_float *A, const int lda,
          const fupermod_float *B, const int ldb,
          const fupermod_float beta,
          fupermod_float *C, const int ldc)
{
  cublasxgemm_params* params = (cublasxgemm_params*)gemm->params;

  // pointer to host workspace
  fupermod_float *hA = params->workspace_WA;
  fupermod_float *hB = params->workspace_WB;

  // if pivot column is local, copy the pivot column to host workspace WA that is allocated as pinned memory
  if (A != hA) {
    int count = sizeof(fupermod_float) * K;
    int i;
    for (i = 0; i < M; i++) {
      memcpy(hA + i * K, A + i * lda, count);
    }
  }
  // if pivot row is local, copy the pivot row to host workspace WB that is allocated as pinned memory
  if (B != hB) {
      int count = sizeof(fupermod_float) * N;
      int i;
      for (i = 0; i < K; i++) {
        memcpy(hB + i * N, B + i * ldb, count);
      }
    } 
/*  if (B != hB) {
    int count = sizeof(fupermod_float) * N;
    int i;
    for (i = 0; i < K; i++) {
      memcpy(params->dma_WB + i * N, B + i * ldb, count);
    }
    hB = params->dma_WB;
  } else {
    if (params->dma_WB) { // when build fpms, dam_WB is NULL
      memcpy(params->dma_WB, params->workspace_WB, sizeof(fupermod_float) * K * N);
      hB = params->dma_WB;
    }
  }*/
  
  // split matrix
  int h_, w_, *height, *width;
  size_t mem_size_device = gama * getDeviceMemory(params->device_id)/sizeof(fupermod_float);
  matrix_split_single_exceed(M, N, K, mem_size_device, &h_, &w_, &height, &width);
  
  // initialization
  if (params->idx_outer == 0) {
    // only one device to be used in this case, so the size of dA, dB, dC is 1
    params->dA = (fupermod_float**)malloc(sizeof(fupermod_float*));
    params->dB = (fupermod_float**)malloc(sizeof(fupermod_float*));
    params->dC = (fupermod_float**)malloc(sizeof(fupermod_float*));
    
    if (PRINT)
      fprintf(stdout, "Kernel: gemm_kernel_gpu_single_exceed_3, M %d N %d K %d h_ %d w_ %d\n", M, N, K, h_, w_);
    CUDA_SAFE_CALL(cudaMalloc((void**)(&params->dA[0]), sizeof(fupermod_float) * height[0] * K * K * 2));
    CUDA_SAFE_CALL(cudaMalloc((void**)(&params->dB[0]), sizeof(fupermod_float) * K * width[0] * K));
    CUDA_SAFE_CALL(cudaMalloc((void**)(&params->dC[0]), sizeof(fupermod_float) * height[0] * K * width[0] * K * 2));
//  CUDA_SAFE_CALL(cudaMemset((void*)params->dC[0], 0,  sizeof(fupermod_float) * height[0] * K * width[0] * K * 2));
  }
  
  //streams and events
  cudaStream_t stream[h_ * w_];
  cudaEvent_t event_a[h_ * w_];
  cudaEvent_t event_b[w_];
  cudaEvent_t event_c[h_ * w_];
  int i, j;
  for (j = 0; j < w_; j++) {
    CUDA_SAFE_CALL(cudaEventCreateWithFlags(&event_b[j], cudaEventDisableTiming));
    for (i = 0; i < h_; i++) {
      if (params->is_overlap == 1) {
        CUDA_SAFE_CALL(cudaStreamCreate(&stream[i + j * h_]));
      } else {// without overlapping
        stream[i + j * h_] = 0;
      }
      CUDA_SAFE_CALL(cudaEventCreateWithFlags(&event_a[i + j * h_], cudaEventDisableTiming));
      CUDA_SAFE_CALL(cudaEventCreateWithFlags(&event_c[i + j * h_], cudaEventDisableTiming));
    }
  }

  int ldda = K;
  fupermod_float *dA = params->dA[0];
  fupermod_float *dB = params->dB[0];
  fupermod_float *dC = params->dC[0];

  // update in an normal order, column by column
  if (params->idx_outer % 2 == 0) {
    for (j = 0; j < w_; j++) {
      for (i = 0; i < h_; i++) {
        int idx = i + j * h_;
        int buf = idx % 2;
        int buf_ = (idx + 1) % 2;
        int i_ = (idx + 1) % h_;
        int j_ = (idx + 1) / h_;
        int lddb = width[j] * K;
        int lddc = width[j] * K;
        
        // initialisation
        if (idx == 0) {
          CUDA_SAFE_CALL(cudaMemcpyAsync(dB, hB, sizeof(fupermod_float) * width[j] * K * K, cudaMemcpyHostToDevice, stream[idx]));
          CUDA_SAFE_CALL(cudaMemcpyAsync(dA, hA, sizeof(fupermod_float) * K * height[i] * K, cudaMemcpyHostToDevice, stream[idx]));
          if (params->idx_outer == 0) {
            CUDA_SAFE_CALL(cudaMemcpyAsync(dC, C, sizeof(fupermod_float) * width[j] * K * height[i] * K, cudaMemcpyHostToDevice, stream[idx]));
          }
        }
        // gemm and send next blocks
        if (idx < (h_ * w_ - 1)) {
          // gemm
          cublasSetStream(params->handles[0], stream[idx]);
          CUBLAS_SAFE_CALL(CUBLAS_GEMM(params->handles[0], trans_cblas2cublas(TransB), trans_cblas2cublas(TransA), width[j] * K, height[i] * K, K,
                        &alpha, dB, lddb, addr(dA,buf,j,height,width,N,K,4), ldda, &beta, addr(dC,buf,j,height,width,N,K,5), lddc));
          CUDA_SAFE_CALL(cudaEventRecord(event_a[idx], stream[idx]));
          if (i == (h_ - 1)) {
            CUDA_SAFE_CALL(cudaEventRecord(event_b[j], stream[idx]));
          }
          // if possible, send next block of B
          if (i == (h_ - 1)) {
            if (j < (w_ - 1)) {
              cudaStreamWaitEvent(stream[idx+1], event_b[j], 0);
            }
            CUDA_SAFE_CALL(cudaMemcpyAsync(dB, addr1D(hB,i_,j_,height,width,h_,w_,K,2), sizeof(fupermod_float) * width[j_] * K * K, cudaMemcpyHostToDevice, stream[idx+1]));
          }
          // send next block of A
          if (idx > 0) {
            cudaStreamWaitEvent(stream[idx+1], event_a[idx-1], 0);
          }
          CUDA_SAFE_CALL(cudaMemcpyAsync(addr1D(dA,buf_,j_,height,width,h_,w_,K,4), addr1D(hA,i_,j_,height,width,h_,w_,K,1), 
                                         sizeof(fupermod_float) * K * height[i_] * K, cudaMemcpyHostToDevice, stream[idx+1]));
          // send next block of C
          if (params->idx_outer == 0) {
            if (idx > 0) {
              cudaStreamWaitEvent(stream[idx+1], event_c[idx-1], 0);
            }
            CUDA_SAFE_CALL(cudaMemcpyAsync(addr1D(dC,buf_,j_,height,width,h_,w_,K,5), addr1D(C,i_,j_,height,width,h_,w_,K,3), 
                                           sizeof(fupermod_float) * width[j_] * K * height[i_] * K, cudaMemcpyHostToDevice, stream[idx+1]));
          } else {
            if (idx > 0) {
              cudaStreamWaitEvent(stream[idx+1], event_c[idx-1], 0);
              CUDA_SAFE_CALL(cudaMemcpyAsync(addr1D(dC,buf_,j_,height,width,h_,w_,K,5), addr1D(C,i_,j_,height,width,h_,w_,K,3), 
                                             sizeof(fupermod_float) * width[j_] * K * height[i_] * K, cudaMemcpyHostToDevice, stream[idx+1]));
            }
          }
          // recv block of C
          if  (params->idx_outer == (params->num_iteration - 1)) {
            CUDA_SAFE_CALL(cudaMemcpyAsync(addr1D(C,i,j,height,width,h_,w_,K,3), addr1D(dC,buf,j,height,width,h_,w_,K,5), 
                      sizeof(fupermod_float) * width[j] * K * height[i] * K, cudaMemcpyDeviceToHost, stream[idx]));
            CUDA_SAFE_CALL(cudaEventRecord(event_c[idx], stream[idx]));
          } else {
            if (idx < (h_ * w_ - 2)) {
              CUDA_SAFE_CALL(cudaMemcpyAsync(addr1D(C,i,j,height,width,h_,w_,K,3), addr1D(dC,buf,j,height,width,h_,w_,K,5), 
                        sizeof(fupermod_float) * width[j] * K * height[i] * K, cudaMemcpyDeviceToHost, stream[idx]));
              CUDA_SAFE_CALL(cudaEventRecord(event_c[idx], stream[idx]));  
            }
          }      
        } else {
          // the last gemm
          cublasSetStream(params->handles[0], stream[idx]);
          CUBLAS_SAFE_CALL(CUBLAS_GEMM(params->handles[0], trans_cblas2cublas(TransB), trans_cblas2cublas(TransA), width[j] * K, height[i] * K, K,
                        &alpha, dB, lddb, addr(dA,buf,j,height,width,N,K,4), ldda, &beta, addr(dC,buf,j,height,width,N,K,5), lddc));
          if (params->idx_outer == (params->num_iteration - 1)) {
            CUDA_SAFE_CALL(cudaMemcpyAsync(addr1D(C,i,j,height,width,h_,w_,K,3), addr1D(dC,buf,j,height,width,h_,w_,K,5), 
                      sizeof(fupermod_float) * width[j] * K * height[i] * K, cudaMemcpyDeviceToHost, stream[idx]));
          }
        }
      }
    }
  } else {
    for (j = (w_ - 1); j >= 0 ; j--) {
      for (i = (h_ - 1); i >= 0 ; i--) {
        int idx = i + j * h_;
        int buf = idx % 2;        
        int buf_ = (idx - 1) % 2;
        int i_ = (idx - 1) % h_;
        int j_ = (idx - 1) / h_;
        int lddb = width[j] * K;
        int lddc = width[j] * K;
                
        // initialisation
        if (idx == (h_ * w_ - 1)) {
          CUDA_SAFE_CALL(cudaMemcpyAsync(dB, addr1D(hB,i,j,height,width,h_,w_,K,2), sizeof(fupermod_float) * width[j] * K * K, cudaMemcpyHostToDevice, stream[idx]));
          CUDA_SAFE_CALL(cudaMemcpyAsync(addr1D(dA,buf,j,height,width,h_,w_,K,4), addr1D(hA,i,j,height,width,h_,w_,K,1), 
                                         sizeof(fupermod_float) * K * height[i] * K, cudaMemcpyHostToDevice, stream[idx]));
        }  
        // gemm and send next blocks
        if (idx > 0) {
          // gemm
          cublasSetStream(params->handles[0], stream[idx]);
          CUBLAS_SAFE_CALL(CUBLAS_GEMM(params->handles[0], trans_cblas2cublas(TransB), trans_cblas2cublas(TransA), width[j] * K, height[i] * K, K,
                        &alpha, dB, lddb, addr(dA,buf,j,height,width,N,K,4), ldda, &beta, addr(dC,buf,j,height,width,N,K,5), lddc));
          CUDA_SAFE_CALL(cudaEventRecord(event_a[idx], stream[idx]));
          if (i == 0) {
            CUDA_SAFE_CALL(cudaEventRecord(event_b[j], stream[idx]));
          }        
          // if possible, send next block of B
          if (i == 0) {
            if (j > 0) {
              cudaStreamWaitEvent(stream[idx-1], event_b[j], 0);
            }
            CUDA_SAFE_CALL(cudaMemcpyAsync(dB, addr1D(hB,i_,j_,height,width,h_,w_,K,2), sizeof(fupermod_float) * width[j_] * K * K, cudaMemcpyHostToDevice, stream[idx-1]));
          }
          // send next block of A
          if (idx < (h_ * w_ - 1)) {
            cudaStreamWaitEvent(stream[idx-1], event_a[idx+1], 0);
          }
          CUDA_SAFE_CALL(cudaMemcpyAsync(addr1D(dA,buf_,j_,height,width,h_,w_,K,4), addr1D(hA,i_,j_,height,width,h_,w_,K,1), 
                                         sizeof(fupermod_float) * K * height[i_] * K, cudaMemcpyHostToDevice, stream[idx-1]));
          // send next block of C
          if (idx < (h_ * w_ - 1)) {
            cudaStreamWaitEvent(stream[idx-1], event_c[idx+1], 0);
            CUDA_SAFE_CALL(cudaMemcpyAsync(addr1D(dC,buf_,j_,height,width,h_,w_,K,5), addr1D(C,i_,j_,height,width,h_,w_,K,3), 
                                           sizeof(fupermod_float) * width[j_] * K * height[i_] * K, cudaMemcpyHostToDevice, stream[idx-1]));
          }
          // recv block of C
          if  (params->idx_outer == (params->num_iteration - 1)) {
            CUDA_SAFE_CALL(cudaMemcpyAsync(addr1D(C,i,j,height,width,h_,w_,K,3), addr1D(dC,buf,j,height,width,h_,w_,K,5), 
                                           sizeof(fupermod_float) * width[j] * K * height[i] * K, cudaMemcpyDeviceToHost, stream[idx]));
            CUDA_SAFE_CALL(cudaEventRecord(event_c[idx], stream[idx]));
          } else {
            if (idx > 1) {
              CUDA_SAFE_CALL(cudaMemcpyAsync(addr1D(C,i,j,height,width,h_,w_,K,3), addr1D(dC,buf,j,height,width,h_,w_,K,5), 
                                             sizeof(fupermod_float) * width[j] * K * height[i] * K, cudaMemcpyDeviceToHost, stream[idx]));
              CUDA_SAFE_CALL(cudaEventRecord(event_c[idx], stream[idx]));  
            }
          }
        } else {
          // the last gemm
          cublasSetStream(params->handles[0], stream[idx]);
          CUBLAS_SAFE_CALL(CUBLAS_GEMM(params->handles[0], trans_cblas2cublas(TransB), trans_cblas2cublas(TransA), width[j] * K, height[i] * K, K,
                        &alpha, dB, lddb, addr(dA,buf,j,height,width,N,K,4), ldda, &beta, addr(dC,buf,j,height,width,N,K,5), lddc));
          if (params->idx_outer == (params->num_iteration - 1)) {
            CUDA_SAFE_CALL(cudaMemcpyAsync(addr1D(C,i,j,height,width,h_,w_,K,3), addr1D(dC,buf,j,height,width,h_,w_,K,5), 
                                           sizeof(fupermod_float) * width[j] * K * height[i] * K, cudaMemcpyDeviceToHost, stream[idx]));
          }
        }
      }
    }
  }
  CUDA_SAFE_CALL(cudaDeviceSynchronize());
  
  if (PRINT)
    fprintf(stdout, "idx_outer %d\n", params->idx_outer);
  params->idx_outer++;

  // destroy events
  for (j = 0; j < w_; j++) {
    CUDA_SAFE_CALL(cudaEventDestroy(event_b[j]));
    for (i = 0; i < h_; i++) {
      if (params->is_overlap == 1) {
        CUDA_SAFE_CALL(cudaStreamDestroy(stream[i + j * h_]));
      }
      CUDA_SAFE_CALL(cudaEventDestroy(event_a[i + j * h_]));
      CUDA_SAFE_CALL(cudaEventDestroy(event_c[i + j * h_]));
    }
  }
  // free memory
  free(height); height = NULL; free(width); width = NULL;
  // at the last iteration, free device memory, and destroy handles
  if ((params->num_iteration) && (params->idx_outer >= params->num_iteration)) {
    if (PRINT)
      fprintf(stdout, "idx_outer %d, params->num_iteration %d, free device memory dA/B/C ooc\n", params->idx_outer, params->num_iteration);
    CUDA_SAFE_CALL(cudaFree(params->dA[0])); free(params->dA); params->dA = NULL;
    CUDA_SAFE_CALL(cudaFree(params->dB[0])); free(params->dB); params->dB = NULL;
    CUDA_SAFE_CALL(cudaFree(params->dC[0])); free(params->dC); params->dC = NULL;
    CUBLAS_SAFE_CALL(cublasDestroy(params->handles[0])); free(params->handles); params->handles = NULL;
  }
}
#endif
