#include "fupermod/fupermod_debug.h"
#include "fupermod/fupermod_benchmark.h"
#include "fupermod/fupermod_dynamic.h"
#include "fupermod_hybrid.h"
#include "mm_col_distr.h"
#include <stdlib.h>
#include <sys/time.h>
#include <errno.h>
#include <math.h>
#include <string.h>

int print_models = 1;
int print_models_data = 1;

int hybrid_node_partitioner_iterate(fupermod_dynamic* partitioner, fupermod_benchmark_hybrid* hybrid);
int hybrid_node_partitioner_one(fupermod_dynamic* partitioner, fupermod_benchmark_hybrid* hybrid);
int hybrid_device_build_full_models(fupermod_dynamic* partitioner, fupermod_benchmark_hybrid* hybrid);

// called by inter-node partitioner.
int fupermod_hybrid_benchmark_execute(struct fupermod_benchmark* benchmark, MPI_Comm comm, long long int d, fupermod_precision precision, fupermod_point* point) {
  NOT_USED(precision);
  fupermod_benchmark_hybrid* hybrid = (fupermod_benchmark_hybrid*)benchmark;
  int rank;
  MPI_Comm_rank(comm, &rank);
  int size; 
  MPI_Comm_size(comm, &size);

  // compute idist.
  double df = d;
  double* areas = (double*)malloc(sizeof(double) * size);
  MPI_Allgather(&df, 1, MPI_DOUBLE, areas, 1, MPI_DOUBLE, comm);

  hybrid->mm_col_dist = mm_col_dist(size, hybrid->params.N / hybrid->params.kb, hybrid->params.kb, areas);
  int myh = hybrid->mm_col_dist->d[rank].h;
  int myw = hybrid->mm_col_dist->d[rank].w;
  // Hack: for building full models.
  //fprintf(stderr, "WARNING-code setup for building ffpm's only!!\n");
  //myh = myw = sqrt(df);
  // Hack: add small point to model: check that all 1.
  int i;
  for (i = 0; i < size; i++) {
    fprintf(stderr, "area %d %lf\n", i, areas[i]); 
    if (areas[i] != 1)
      break;
  }
  if (i == size) {
    myh = myw = 1;
    fprintf(stderr, "### Info: running benchmark size 1\n");
  }
  if (myh == 0 || myw == 0) {
    point->d = myh * myw;
    point->t = -1;
    point->reps = 1;
    point->ci = 0;
    return 0;
  }

  fprintf(stderr, "Running %s\n", __FUNCTION__);
  // initialise
  void* ptr = &hybrid->params;
  // Dave, please handle params in the kernel initialize - you can get myh and myw from prt
  int rc = benchmark->kernel->initialize(myh * myw, &ptr);
  if (rc == FUPERMOD_ALLOC_FAIL) {
    point->d = myh * myw;
    point->t = FUPERMOD_MAX_TIME;
    point->reps = 0;
    point->ci = 0;
    return rc;
  }
  if (rc != FUPERMOD_SUCCESS) {
    fprintf(stderr, "%s: kernel initialization failed at proc %d\n", __func__, rank);
    return -1;
  }
  int kb = hybrid->params.kb;
  int rc_hybrid_partition = fupermod_hybrid_partition_node(hybrid, myw * kb);

  point->d = myh * myw;
  point->t = hybrid->params.time;
  point->reps = 1;
  point->ci = 0;
  // finalise
  if (benchmark->kernel->finalize(NULL) != FUPERMOD_SUCCESS) {
    fprintf(stderr, "%s: kernel finalization failed at proc %d\n", __func__, rank);
    return -1;
  }
  if (rc_hybrid_partition == FUPERMOD_ALLOC_FAIL) {
    fprintf(stderr, "Warning: Returning ALLOC_FAIL %s\n", __func__);
    return rc_hybrid_partition;
  }
  return 0;
}

// partition within node
int fupermod_hybrid_partition_node(struct fupermod_benchmark_hybrid* hybrid, int N){
  //	fupermod_benchmark_hybrid* hybrid = (fupermod_benchmark_hybrid*)benchmark;
  int node_size = hybrid->params.hybridconf.numGPUs + hybrid->params.hybridconf.numCPUs;
  fupermod_data** node_data = (fupermod_data**)malloc(sizeof(fupermod_data*) * node_size);
  fupermod_model** node_models = (fupermod_model**)malloc(sizeof(fupermod_model*) * node_size);

  int i;
  for (i = 0; i < node_size; i++) {
    node_data[i] = fupermod_data_alloc();
    node_models[i] = hybrid->partitioner.model_alloc(node_data[i], hybrid->base.kernel->complexity, N, NULL, hybrid->partitioner.arg);
  }

  fupermod_dynamic partitioner_dyn = {hybrid->partitioner.partition, node_size, node_models, fupermod_dist_alloc(N, node_size)};

  int iter = 0;
  int status = 0;
  // innter loop
  // If status is: 
  // 0: benchmark worked and should continue iterating, 
  // less than 0 iteration should not be incremented. 
  // Greater than 0: distribution has converged.

  // adds a benchmark with dist of 1 to models. 
  hybrid_node_partitioner_one(&partitioner_dyn, hybrid);

  while (iter < 50) {
    fprintf(stderr, "Info: Starting node iter:%d of %d\n", iter, hybrid->iters);
    status = hybrid_node_partitioner_iterate(&partitioner_dyn, hybrid);
    iter++;
    if (iter >= hybrid->iters && status == 0)
      break;
    if (status > 0)
      break;
    if (hybrid->iters == 2 && iter >= hybrid->iters) //special case for CPM1 - break even if status isn't 0.
      break;
  }
  //	fprintf(stderr, "Warning: not doing usual thing! building device models with hack!\n");
  //	hybrid_device_build_full_models(&partitioner_dyn, hybrid);

  // Print models
  if (print_models) { 
    //todo fix.
    /*
    //Out of date:
    char filename[256];
    char hostname[MPI_MAX_PROCESSOR_NAME];
    int len;
    MPI_Get_processor_name(hostname, &len);
    int pos = strcspn(hostname, ".");
    sprintf(filename, "models.%.*s.dat", pos, hostname);
    FILE* file = fopen(filename, "w");
    fprintf(file, "#N:%d host:%s\n", N, hostname);
    fupermod_print_models(file, node_size, node_models, N);
    fclose(file);
    //
    //new version from partitioner.c
    for (i = 0; i < size; i++) {
      char* filename = fupermod_model_filename(data_dir, confs[i]);
      FILE* outfile = fopen(filename, "w");
      free(filename); filename = NULL;
      if (outfile == NULL)
        MPI_Abort(comm, -1);
      fupermod_print_model(outfile, library, confs[i], models[i], D);
      fclose(outfile);
    }
    */
  }
  if (print_models_data) {
    char filename[256];
    char hostname[MPI_MAX_PROCESSOR_NAME];
    int len;
    MPI_Get_processor_name(hostname, &len);
    int pos = strcspn(hostname, ".");
    sprintf(filename, "points.%.*s.dat", pos, hostname);
    FILE* file = fopen(filename, "w");
    fprintf(file, "#N:%d host:%s\n", N, hostname);
    fupermod_print_models_data(file, node_size, node_models);
    fclose(file);
  }

  fupermod_dist_free(partitioner_dyn.dist);
  for (i = 0; i < node_size; i++) {
    fupermod_data_free(node_data[i]);
    hybrid->partitioner.model_free(node_models[i]);
  }
  free(node_data);
  free(node_models);
  return status;
}

fupermod_benchmark_hybrid* fupermod_benchmark_hybrid_alloc(fupermod_kernel* kernel, HybridConfig* config, int N, int kb, int iters, double eps, fupermod_partitioner partitioner) {
  //hybridconf = config;
  fupermod_benchmark_hybrid* hybrid = (fupermod_benchmark_hybrid*)malloc(sizeof(fupermod_benchmark_hybrid));
  hybrid->base.kernel = kernel;
  //  TODO: set to function which will get conf
//  hybrid->base.conf = FUPERMOD_DEFAULT_DEVICE; 
  hybrid->base.execute = fupermod_hybrid_benchmark_execute;
  hybrid->params.hybridconf = *config;
  hybrid->params.time = 0;
  hybrid->params.kb = kb;
  hybrid->params.N = N;
  hybrid->iters = iters;
  hybrid->eps = eps;
  hybrid->partitioner = partitioner;
  return hybrid;
}

void fupermod_benchmark_hybrid_free(fupermod_benchmark_hybrid* benchmark) {
  free(benchmark);
}


int hybrid_node_partitioner_iterate(fupermod_dynamic* partitioner, fupermod_benchmark_hybrid* hybrid) {
  fupermod_point* points = (fupermod_point*)malloc(sizeof(fupermod_point) * partitioner->size);
  int i;
  for (i = 0; i < partitioner->size; i++){
    points[i].d = hybrid->params.hybridconf.deviceWorkload[i] = partitioner->dist->parts[i].d;
  }
  fprintf(stderr, "N: %lld Nodedist: ", partitioner->dist->D);
  for (i = 0; i < partitioner->size; i++)
    fprintf(stderr, "%d ", hybrid->params.hybridconf.deviceWorkload[i]);
  fprintf(stderr, "\n");
  // execute
  // status may be: 
  // 	0 successfull execution
  //	-1 failled exectuion
  //	-2 failled malloc 
  int status = hybrid->base.kernel->execute(NULL, &hybrid->params);
  if (status == -1) {
    fprintf(stderr, "Execution of kernel failed, returned -1.  %s\n", __func__);
    MPI_Abort(MPI_COMM_WORLD, 70);
  }

  for (i = 0; i < partitioner->size; i++){
    points[i].t = hybrid->params.hybridconf.deviceTime[i];
    points[i].reps = 1;
    points[i].ci = 0;
    partitioner->models[i]->update(partitioner->models[i], points[i]);
  }
  if (status == 0) 
    status = fupermod_dist_test_times(points, partitioner->size, hybrid->eps);
  // status is now 1 if should finish because error is small
  fprintf(stderr, "times: ");
  for (i = 0; i < partitioner->size; i++)
    fprintf(stderr, "%lf ", points[i].t);
  fprintf(stderr, "\n");

  fupermod_dist* new_dist = fupermod_dist_alloc(partitioner->dist->D, partitioner->size);
  partitioner->partition(new_dist, partitioner->size, partitioner->models);

  // did the dist change? 
  if (status == 0) {
    status = 1;
    for (i = 0; i < partitioner->size; i++) {
      if (partitioner->dist->parts[i].d != new_dist->parts[i].d){
        status = 0;
        break;
      }
    }
  }
  fupermod_dist_free(partitioner->dist);
  partitioner->dist = new_dist;

  if (status) {
    for (i = 0; i < partitioner->size; i++)
      fprintf(stdout, "%lf ", points[i].t);
    fprintf(stdout, "\n");
  }

  free(points); points = NULL;
  // If status is: 0: benchmark worked and should continue iterating, less than 0 iteration should not be incremented. Greater than 0: distribution has converged.
  return status;
}

int hybrid_node_partitioner_one(fupermod_dynamic* partitioner, fupermod_benchmark_hybrid* hybrid) {
  fupermod_point* points = (fupermod_point*)malloc(sizeof(fupermod_point) * partitioner->size);
  int i;
  for (i = 0; i < partitioner->size; i++){
    //points[i].d = hybrid->params.hybridconf.deviceWorkload[i] = partitioner->dist->parts[i].d;
    points[i].d = hybrid->params.hybridconf.deviceWorkload[i] = 1;
  }
  fprintf(stderr, "N: %lld Nodedist: ", partitioner->dist->D);
  for (i = 0; i < partitioner->size; i++)
    fprintf(stderr, "%d ", hybrid->params.hybridconf.deviceWorkload[i]);
  fprintf(stderr, "\n");
  int status = hybrid->base.kernel->execute(NULL, &hybrid->params);
  if (status == -1) {
    fprintf(stderr, "Execution of kernel failed, returned -1.  %s\n", __func__);
    MPI_Abort(MPI_COMM_WORLD, 70);
  }

  for (i = 0; i < partitioner->size; i++){
    points[i].t = hybrid->params.hybridconf.deviceTime[i];
    points[i].reps = 1;
    points[i].ci = 0;
    partitioner->models[i]->update(partitioner->models[i], points[i]);
  }

  free(points); points = NULL;
  return 0;
}

int hybrid_device_build_full_models(fupermod_dynamic* partitioner, fupermod_benchmark_hybrid* hybrid) {
  fupermod_point* points = (fupermod_point*)malloc(sizeof(fupermod_point) * partitioner->size);
  int size = partitioner->size;
  long long int N = partitioner->dist->D;
  long long int G;// to gpu
  int step = 160;
  for (G = 0; G <= N; G+= step){
    long long int R; //remainder to cpu.
    R = N - G;
    points[0].d = hybrid->params.hybridconf.deviceWorkload[0] = G;
    int i;
    for (i = 1; i < partitioner->size; i++){
      points[i].d = hybrid->params.hybridconf.deviceWorkload[i] = R / (size - 1) + (((i-1) < R % (size-1)) ? 1 : 0);
    }
    fprintf(stderr, "N: %lld Nodedist: ", partitioner->dist->D);
    for (i = 0; i < partitioner->size; i++)
      fprintf(stderr, "%d ", hybrid->params.hybridconf.deviceWorkload[i]);
    fprintf(stderr, "\n");

    int status = hybrid->base.kernel->execute(NULL, &hybrid->params);
    if (status == -1) {
      fprintf(stderr, "Execution of kernel failed, returned -1.  %s\n", __func__);
      MPI_Abort(MPI_COMM_WORLD, 70);
    }

    for (i = 0; i < partitioner->size; i++){
      points[i].t = hybrid->params.hybridconf.deviceTime[i];
      points[i].reps = 1;
      points[i].ci = 0;
      partitioner->models[i]->update(partitioner->models[i], points[i]);
    }
  }
  free(points); points = NULL;
  return 0;
}
