#include "fupermod_hybrid.h"
#include <string.h>
#include <stdlib.h>

int hybrid_nodetype_read(char* nodetype_path, HybridConfig* hybridconf) {
  char filename[256];
  char hostname[MPI_MAX_PROCESSOR_NAME];
  int len;
  MPI_Get_processor_name(hostname, &len);
  int pos = strcspn(hostname, ".");
  sprintf(filename, "%s.%.*s.dat", nodetype_path, pos, hostname);
  FILE* stream = NULL;
  if ((stream = fopen(filename, "r")) != NULL) {
    if (fscanf(stream, "#EnvType %d\n#GPU %d\n#CPU %d\n#verbose %d\n", 
          (int*)&hybridconf->EnvType, 
          &hybridconf->numGPUs,	
          &hybridconf->numCPUs, &hybridconf->verbose) != 4) {
      fprintf(stderr, "Error reading from nodediscriptor file: %s\n"
          "Should be like:\n\n"
          "#EnvType 2\n#GPU 0\n#CPU 4\n#verbose 0\n\n", filename);
      return -1;
    }
    fclose(stream);
  } else {
    fprintf(stderr, "Error: Could not open nodediscriptor file: %s\n", filename);
    return -1;
  }
  fprintf(stderr, "#EnvType %d\n#GPU %d\n#CPU %d\n", (int)hybridconf->EnvType, 
      hybridconf->numGPUs,	
      hybridconf->numCPUs);
  int node_size = hybridconf->numGPUs + hybridconf->numCPUs;
  hybridconf->deviceWorkload = (int*)malloc(sizeof(int) * node_size);
  hybridconf->deviceTime = (double*)malloc(sizeof(double) * node_size);
  hybridconf->iterations = 1;
  return 0;
}

int hybrid_nodedist_write(char* nodedist_path, HybridConfig* hybridconf) {
  int node_size = hybridconf->numGPUs + hybridconf->numCPUs;
  char filename[256];
  char hostname[MPI_MAX_PROCESSOR_NAME];
  int len;
  MPI_Get_processor_name(hostname, &len);
  int pos = strcspn(hostname, ".");
  sprintf(filename, "%s.%.*s.dist", nodedist_path, pos, hostname);
  FILE* stream = NULL;
  if ((stream = fopen(filename, "w")) != NULL) {
    int i;
    for (i = 0; i < node_size; i++)
      fprintf(stream, "%d ", hybridconf->deviceWorkload[i]);
    fprintf(stream, "\n");
    fclose(stream);
  } else {
    fprintf(stderr, "Error: Could not open node dist file: %s\n", filename);
    return -1;
  }
  return 0;
}

int hybrid_nodedist_read(char* nodedist_path, HybridConfig* hybridconf) {
  int node_size = hybridconf->numGPUs + hybridconf->numCPUs;
  char filename[256];
  char hostname[MPI_MAX_PROCESSOR_NAME];
  int len;
  MPI_Get_processor_name(hostname, &len);
  int pos = strcspn(hostname, ".");
  sprintf(filename, "%s.%.*s.dist", nodedist_path, pos, hostname);
  FILE* stream = NULL;
  if ((stream = fopen(filename, "r")) != NULL) {
    int i;
    for (i = 0; i < node_size; i++)
      if (fscanf(stream, "%d", &hybridconf->deviceWorkload[i]) != 1) {
        fprintf(stderr, "Error reading %d'th element from node dist file: %s\n", i, filename);
        return -1;
      }
    fclose(stream);
  } else {
    fprintf(stderr, "Error: Could not open node dist file: %s\n", filename);
    return -1;
  }
  {
    fprintf(stderr, "Node %.*s dist:", pos, hostname);
    int i;
    for (i = 0; i < node_size; i++)
      fprintf(stderr, "%d ", hybridconf->deviceWorkload[i]);
    fprintf(stderr, "\n");
  }
  return 0;
}
