#ifndef FUPERMOD_HYBRID_H_
#define FUPERMOD_HYBRID_H_
#include "fupermod/fupermod_dynamic.h"
#include "mm_col_distr.h"

#ifdef __cplusplus
extern "C" {
#endif

#define ROUNDUP 1.1
#define HYBRID_MEMORY 0

#ifdef HAVE_LIBDGEMMHYB
#include <dgemm_hybrid.h>
#include <memory_hybrid.h>
#else
typedef enum { CPU_ONLY, GPU_ONLY, HYBRID } environmentType;

typedef enum { PAGEABLE, PINNED, PINNEDWC } memoryMode;

typedef struct call_config {
  environmentType EnvType;
  int numGPUs;
  int numCPUs;
  int *deviceWorkload;
  double *deviceTime;
  int iterations;
  int verbose;
} HybridConfig;

typedef struct mxm_col_params {
  HybridConfig hybridconf;
  int N;
  int kb;
  double time;
} mxm_col_params;


typedef struct fupermod_benchmark_hybrid {
  fupermod_benchmark base;
  mxm_col_params params;
  mm_col_i_distrib* mm_col_dist;
  fupermod_partitioner partitioner;
  int iters;
  double eps;
} fupermod_benchmark_hybrid;
#endif /* HAVE_LIBDGEMMHYB */

int hybrid_nodedist_read(char* nodedist_path, HybridConfig* hybridconf);
int hybrid_nodedist_write(char* nodedist_path, HybridConfig* hybridconf);
int hybrid_nodetype_read(char* nodetype_path, HybridConfig* hybridconf);
int fupermod_hybrid_partition_node(struct fupermod_benchmark_hybrid* hybrid, int N);
int fupermod_hybrid_partition_node_2D(struct fupermod_benchmark* benchmark, int M, int N, fupermod_precision precision, HybridConfig* config);

/*!
 * Allocates the benchmark for hybrid partitioning */
fupermod_benchmark_hybrid* fupermod_benchmark_hybrid_alloc(fupermod_kernel* kernel, HybridConfig* config, int N, int kb, int iters, double eps, fupermod_partitioner partitioner);

/*! Deallocates the hybrid benchmark */
void fupermod_benchmark_hybrid_free(fupermod_benchmark_hybrid* benchmark);

#ifdef __cplusplus
}
#endif

#endif /* FUPERMOD_HYBRID_H_ */
