#include <stdio.h>
#include <stdlib.h>
#include "fupermod/fupermod.h"
#include "mm_col_distr.h"
#include <string.h>

#define MM_COL_VERBOSE 0

typedef struct tfunction {
  float perimeter;
  int cut;
} tfunction;

typedef struct proc_info {
  double speed;
  int pid;
} proc_info;


typedef struct f_zone {
  float x;
  float y;
  float w;
  float h;
  int i;
  int j;
} f_zone;

typedef struct f_distrib {
  int nproc;
  f_zone *d;
} f_distrib;			
typedef f_distrib *f_distribution;	

int compare_speeds(const void *i, const void *j)
{
  if (((proc_info *) i)->speed > ((proc_info *) j)->speed)
    return (1);
  if (((proc_info *) i)->speed < ((proc_info *) j)->speed)
    return (-1);
  return (0);
}

void minimum_perimeter_sum(proc_info * tab, f_distribution fdist)
{
  int C, q, Copt = 0;
  int r, finished = 0;
  int p = fdist->nproc;

  double S = 0, perimeter;
  double MAXPERI = 2 * p;
  tfunction *tab_f = (tfunction *) malloc(p * p * sizeof(tfunction));


#define f(C,q) tab_f[(C-1) * p + (q-1)]
#define s(i) tab[i-1].speed
#define pid(i) tab[i-1].pid

  qsort(tab, p, sizeof(proc_info), compare_speeds);

  for (q = 1; q <= p; q++) {
    S += s(q);
    f(1, q).perimeter = 1 + S * q;
    f(1, q).cut = 0;
  }

  for (C = 2; (C <= p) && !(finished); C++) {
    for (q = C; q <= p; q++) {
      S = 0;
      f(C, q).perimeter = MAXPERI;
      for (r = q - 1; r >= (C - 1); r--) {
        S += s(r + 1);
        perimeter = 1 + S * (q - r) + f(C - 1, r).perimeter;
        if (perimeter < f(C, q).perimeter) {
          f(C, q).perimeter = perimeter;
          f(C, q).cut = r;
        }
      }
    }
    if (f(C, p).perimeter > f(C - 1, p).perimeter) {
      finished = 1;
      Copt = C - 1;
    }
  }

  if (!finished)
    Copt = p;


  {
    int i;
    float xmin, ymin, w;

    q = p;
    xmin = 0;
    int j_proc = 0;
    for (C = Copt; C >= 1; C--) {
      ymin = 0;
      int i_proc = 0;
      w = 0;
      for (i = q; i > f(C, q).cut; i--)
        w = w + (s(i));

      for (i = q; i > f(C, q).cut; i--) {
        fdist->d[pid(i)].x = xmin;
        fdist->d[pid(i)].y = ymin;
        fdist->d[pid(i)].w = w;
        if(w > 0)
          fdist->d[pid(i)].h = (s(i)) / w;
        else 
          fdist->d[pid(i)].h = 0;
        fdist->d[pid(i)].i = i_proc++;
        fdist->d[pid(i)].j = j_proc;
        if(w > 0)
          ymin += s(i) / w;
      }
      xmin += w;
      j_proc++;
      q = f(C, q).cut;
    }
  }
#undef f
#undef s
#undef pid
  free(tab_f); tab_f = NULL;
}

#define ROUND(x) ((int)((x)+SIGN(x)*0.5))
#define SIGN(x) ((x)>0 ? (1) : (-1))

mm_col_i_distrib* float_to_int_dist(f_distribution fdist, int N, int M, int kb) {

  mm_col_i_distrib* idist = (mm_col_i_distrib*)calloc(sizeof(mm_col_i_distrib),1);
  idist->d = (mm_col_i_zone*) malloc(fdist->nproc * sizeof(mm_col_i_zone));
  idist->nproc = fdist->nproc;
  idist->N = N;
  idist->M = M;
  idist->kb = kb;

  int i;
  for (i = 0; i < fdist->nproc; i++) {
    idist->d[i].x = ROUND((fdist->d[i].x) * N);
    idist->d[i].w = ROUND((fdist->d[i].w + fdist->d[i].x) * N) - idist->d[i].x;
    idist->d[i].y = ROUND((fdist->d[i].y) * M);
    idist->d[i].h = ROUND((fdist->d[i].h + fdist->d[i].y) * M) - idist->d[i].y;
    idist->d[i].i = fdist->d[i].i;
    idist->d[i].j = fdist->d[i].j;
  }

  if(MM_COL_VERBOSE) {
    printf("\n#i\tx\ty\tw\th\tw*h\n");
    for (i = 0; i < fdist->nproc; i++) {
      printf("%d\t%d\t%d\t%d\t%d\t%d\n", i, 
          idist->d[i].x, 
          idist->d[i].y,
          idist->d[i].w,
          idist->d[i].h,
          idist->d[i].w * idist->d[i].h);
    }
  }

  // Check everything adds up.
  int	max_j = 0;
  for (i = 0; i < fdist->nproc; i++) {
    if (idist->d[i].j > max_j) {
      max_j = idist->d[i].j;
    }
  }
  int *h_tot = (int*)malloc(sizeof(int)*(max_j+1));
  int *col_widths = (int*)malloc(sizeof(int)*(max_j+1));
  for (i = 0; i <= max_j; i++) {
    h_tot[i] = 0;
    col_widths[i] = -1;
  }
  int w_tot = 0;
  for (i = 0; i < fdist->nproc; i++) {
    if (idist->d[i].i == 0) {
      // add up all column heads.
      w_tot += idist->d[i].w;
      col_widths[idist->d[i].j] = idist->d[i].w;
    }
    h_tot[idist->d[i].j] += idist->d[i].h;
    if (idist->d[i].x < 0 ||
        idist->d[i].y < 0 ||
        idist->d[i].x + idist->d[i].w > N ||
        idist->d[i].y + idist->d[i].h > M) {
      fprintf(stderr, "Error, dist out of bounds\n");
      fprintf(stderr, "Error: x:%d y:%d w:%d h:%d N:%d M:%d\n", idist->d[i].x, idist->d[i].y, idist->d[i].w, idist->d[i].h, N, M);
      // exit(-1);
    }	
  }
  if (w_tot != N) {
    fprintf(stderr, "Error, widths do not add up, will attempt to fix\n");
    w_tot = N - w_tot;
    for (i = 0; i < fdist->nproc; i++) {
      if (idist->d[i].j == max_j) {
        idist->d[i].w += w_tot;
      }
    }
  }
  int j;
  //	fprintf(stderr, "max_j:%d\n",max_j);
  for (j = 0; j <= max_j; j++) {
    //		fprintf(stderr, "h_tot[%d]=%d\n",j, h_tot[j]);
    if (h_tot[j] != N && col_widths[j] != 0) {
      fprintf(stderr, "Error, heights in col%d do not add up, will attempt to fix\n", j);
      int max_i = 0;
      for (i = 0; i < fdist->nproc; i++) {
        if (idist->d[i].j == j && idist->d[i].i > max_i) {
          max_i = idist->d[i].i;
        }
      }
      for (i = 0; i < fdist->nproc; i++) {
        if (idist->d[i].j == j && idist->d[i].i == max_i) {
          //					printf("i:%d j%d dif:%d\n",i,j, (N - h_tot[j]));
          idist->d[i].h += (N - h_tot[j]);
        }
      }
      // now fix y coords
      int _i = 0, y = 0;
      for (_i = 0; _i <= max_i; _i++) {
        for (i = 0; i < fdist->nproc; i++) {
          if (idist->d[i].j == j && idist->d[i].i == _i) {
            idist->d[i].y = y;
            y += idist->d[i].h;
          }
        }
      }
    }
  }
  free(h_tot);
  free(col_widths);

  return (idist);
}


mm_col_i_distrib* mm_col_dist(int nproc, int N, int kb, double* speeds) {
  // float distribution, distributing to unit square.
  f_distribution dist = calloc(sizeof(f_distrib),1);
  dist->d = malloc(nproc * sizeof(f_zone));
  dist->nproc = nproc;

  proc_info *tab = malloc(nproc * sizeof(proc_info));
  double sum_s = 0.0;
  int i;
  for (i = 0; i < nproc; i++) {
    tab[i].pid = i;
    tab[i].speed = speeds[i];
    sum_s += tab[i].speed;
  }
  // Normalise speeds
  for (i = 0; i < nproc; i++) {
    tab[i].speed /= sum_s;
  }

  // Calculate the optimum shapes which will minimise communication
  minimum_perimeter_sum(tab, dist);
  if(MM_COL_VERBOSE) {
    int kb = 32; // guess for printing memory!
    printf("#i\tx       \ty       \tl       \th       \tl*h     \tmemory (if kb=32)\n");
    for (i = 0; i < nproc; i++) {
      printf("%d\t%f\t%f\t%f\t%f\t%f\t%fMB\n", i,  
          dist->d[i].x, 
          dist->d[i].y,
          dist->d[i].w,
          dist->d[i].h,
          dist->d[i].w *	dist->d[i].h, 
          8 * (3.0 * dist->d[i].w *	dist->d[i].h * kb * kb +  kb * dist->d[i].w + kb * dist->d[i].h)/1024/1024);
    }
  }

  // convert unit square float dist to full size int dist.
  mm_col_i_distrib* id = float_to_int_dist(dist, N, N, kb);

  free(dist->d);
  free(dist); dist = NULL;
  free(tab); tab = NULL;

  return id;
}

void mm_col_i_dist_free(mm_col_i_distrib* dist) {
  free(dist->d);
  free(dist); dist = NULL;
}

void mm_col_print(mm_col_i_distrib* dist) {
  int kb = dist->kb;
  int i;
  printf("\n#\ti\tj\tx\ty\tw\th\tw*h\tmemory\n");
  for (i = 0; i < dist->nproc; i++) {
    printf("%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%dMB\n", i, 
        dist->d[i].i, 
        dist->d[i].j,
        dist->d[i].x, 
        dist->d[i].y,
        dist->d[i].w,
        dist->d[i].h,
        dist->d[i].w * dist->d[i].h,
        (int)(8 * (3.0 * dist->d[i].w *	dist->d[i].h * kb * kb +  kb * dist->d[i].w + kb * dist->d[i].h)/1024/1024));
  }
}

int mm_col_dist_write(FILE* stream, mm_col_i_distrib* idist, fupermod_dist* dist, fupermod_process_conf* confs) {
  int kb = idist->kb;
  fprintf(stream, "#size %d\n", idist->nproc);
  fprintf(stream, "#N %d\n", idist->N);
  fprintf(stream, "#kb %d\n", kb);
  fprintf(stream, "%-15s  rank_intra\ti\tj\tx\ty\tw\th\tw*h\tmem\ttime\n", "#host");
  int i;
  for (i = 0; i < idist->nproc; i++) {
    double t = 0;
    if(dist!=NULL) t = dist->parts[i].t;
    fprintf(stream, "%-20s\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%lld\t%dMB\t%lf\n",
        confs[i].hostname,
        confs[i].rank_intra,
        idist->d[i].i, 
        idist->d[i].j,
        idist->d[i].x, 
        idist->d[i].y,
        idist->d[i].w,
        idist->d[i].h,
        (long long int)idist->d[i].w * idist->d[i].h,
        (int)(8 * (3.0 * idist->d[i].w *	idist->d[i].h * kb * kb +  kb * idist->d[i].w + kb * idist->d[i].h)/1024/1024),
        t
        );
  }
  return 0;
}

mm_col_i_distrib* mm_col_dist_read(FILE* stream, int size, fupermod_process_conf* confs) {
  int size_read = 0;
  /* N is matrix dimension in kb blocks. Total matrix size is N*kb x N*kb. */
  int N = 0;
  int kb = 0;
  fscanf(stream, "#size %d\n", &size_read);
  if (size_read != size) {
    fprintf(stderr, "Error: Size read is != size in %s\n", __func__);
    return NULL;
  }

  mm_col_i_distrib* dist = (mm_col_i_distrib*)malloc(sizeof(mm_col_i_distrib));
  dist->d = (mm_col_i_zone*)malloc(size * sizeof(mm_col_i_zone));

  fscanf(stream, "#N %d\n", &N);
  if (N <= 0) {
    fprintf(stderr, "Error: Reading problem size N\n");
    return NULL;
  }
  fscanf(stream, "#kb %d\n", &kb);
  if (N <= 0) {
    fprintf(stderr, "Error: Reading kb\n");
    return NULL;
  }
  dist->M = dist->N = N;
  dist->nproc = size;
  dist->kb = kb;

  int* check = (int*)calloc(sizeof(int), size);

  int i;
  char *line = NULL;
  size_t len = 0;
  int read;
  getline(&line, &len, stream); //sikp column headers
  for (i = 0; i < size; i++){
    char host[MPI_MAX_PROCESSOR_NAME];
    int rank_intra;
    int i_, j_, x, y, w, h;
    if ((read = getline(&line, &len, stream)) != -1) {
      sscanf(line, "%s\t%d\t%d\t%d\t%d\t%d\t%d\t%d", host, &rank_intra, &i_, &j_, &x, &y, &w, &h);
    } else {
      fprintf(stderr, "Error reading line\n");
      fprintf(stderr, "Line %d: %s\n", i, line);
      return NULL;
    }
    int j;
    for (j = 0;; j++){
      if (j == size){
        fprintf(stderr, "Error: Hostname in dist file not matched with hostname in this MPI instance\n");
        fprintf(stderr, "Line %d: %s\n", i, line);
        return NULL;
      }
      if (check[j] == 0 && strcmp(host, confs[j].hostname) == 0 && confs[j].rank_intra == rank_intra) {
        check[j] = 1;
        dist->d[j].i = i_;
        dist->d[j].j = j_;
        dist->d[j].x = x;
        dist->d[j].y = y;
        dist->d[j].w = w;
        dist->d[j].h = h;
        break;
      }
    }
  }
  // Now zero remaining nodes
  for (i = 0; i < size; i++){
    if (check[i] == 0) {
      check[i] = 1;
      dist->d[i].i = 0;
      dist->d[i].j = 0;
      dist->d[i].x = 0;
      dist->d[i].y = 0;
      dist->d[i].w = 0;
      dist->d[i].h = 0;
    }
  }
  free(check);
  return dist;	
}

