#ifndef MM_COL_DISTR_H
#define MM_COL_DISTR_H

#include <stdio.h>
#include "fupermod/fupermod_partition.h"

#ifdef __cplusplus
extern "C" {
#endif

/*! structure describing size and coordinates 
 * of elements assigned to a single processor  */
typedef struct _mm_col_i_zone {
  /*! horizontal position */
  int x;
  /*! vertical position */
  int y;
  /*! width */
  int w;
  /*! height */
  int h;
  /*! i'th proc in column */
  int i;
  /*! proc in j'th column */
  int j;
  /*! for constructing communicator */
  int c;
} mm_col_i_zone;

/*! Column-wise 2D partitioning */
typedef struct _mm_col_i_distrib {
  /*! number of processors */
  int nproc;
  /*! array of mm_col_i_zone elements of size nproc */
  mm_col_i_zone* d;
  /*! width of total blocked problem size */
  int N;
  /*! height of total blocked problem size */
  int M;
  /*! blocking factor */
  int kb;
} mm_col_i_distrib;

/*!
 * Column-wise 2D partitioning which minimises communication.
 * \param nproc Total number of processors
 * \param N Blocked dimension of matrix
 * \param speeds speeds of processors (Absolute or Relative)
 * \return \ref mm_col_i_distribution structure which holds the position and size of each processor.
 */
mm_col_i_distrib* mm_col_dist(int nproc, int N, int kb, double* speeds);

/*! Deallocate the column-wise 2D partitioning */
void mm_col_i_dist_free(mm_col_i_distrib* dist);

/*! Print out the column-wise 2D partitioning */
void mm_col_print(mm_col_i_distrib* dist);

/*! Output the distribution with hostnames for later reading in */
int mm_col_dist_write(FILE* stream, mm_col_i_distrib* idist, fupermod_dist* dist, fupermod_process_conf* confs);

/*! readin the distribution from file */
mm_col_i_distrib* mm_col_dist_read(FILE* stream, int size, fupermod_process_conf* confs);

#ifdef __cplusplus
}
#endif

#endif /* MM_COL_DISTR_H */
