#ifndef DOXYMAIN_H_
#define DOXYMAIN_H_

/*!
  \mainpage Introduction
  Optimal distribution of computations over the processors of a heterogeneous network
  can be only achieved based on a performance model of the processors which is able to
  accurately predict the contribution of computations into the total execution time of
  any particular parallel or distributed algorithm on the network. The project proposes
  and investigates realistic performance models of processors in heterogeneous networks,
  taking into account all the features that may have an impact on the performance,
  such as architectural and platform heterogeneity of the processors, memory heterogeneity,
  different levels of integration into the network. The project looks at different aspects
  of the models including accuracy and adequacy, efficiency of their use, building,
  maintenance and runtime update.
 
  \latexonly\cite{97}\endlatexonly
 
  \section authors Authors
        <div class="field-items">
              <div class="field-item odd">
                    <a href="http://hcl.ucd.ie/user/alexey-lastovetsky" title="View user profile.">Alexey Lastovetsky</a>        </div>
              <div class="field-item even">
                    <a href="http://hcl.ucd.ie/user/ravi-reddy" title="View user profile.">Ravi Reddy</a>        </div>
              <div class="field-item odd">
                    <a href="http://hcl.ucd.ie/user/robert-higgins" title="View user profile.">Robert Higgins</a>        </div>
              <div class="field-item even">
                    <a href="http://hcl.ucd.ie/user/vladimir-rychkov" title="View user profile.">Vladimir Rychkov</a>        </div>
              <div class="field-item odd">
                    <a href="http://hcl.ucd.ie/user/david-clarke" title="View user profile.">David Clarke</a>        </div>
              <div class="field-item even">
                    <a href="http://hcl.ucd.ie/user/zhong-ziming" title="View user profile.">Zhong Ziming</a>        </div>
              <div class="field-item odd">
                    <a href="http://hcl.ucd.ie/user/jean-no%C3%ABl-quintin" title="View user profile.">Jean-Noël Quintin</a>        </div>
              <div class="field-item even">
                    <a href="http://hcl.ucd.ie/user/khalid-hasanov" title="View user profile.">Khalid Hasanov</a>        </div>
              <div class="field-item odd">
                    <a href="http://hcl.ucd.ie/user/tania-malik" title="View user profile.">Tania Malik</a>        </div>
              <div class="field-item even">
                    <a href="http://hcl.ucd.ie/user/ken-obrien" title="View user profile.">Ken O'Brien</a>        </div>
        </div>
 
  Heterogeneous Computing Laboratory\n
  School of Computer Science and Informatics, University College Dublin\n
  Belfield, Dublin 4, Ireland\n
  http://hcl.ucd.ie
 
  \latexonly
  \bibliographystyle{plain}
  \bibliography{fupermod}
  \endlatexonly
 
  \page installation Installation
  \include README
 
  \page start Quick Start Guide
  <b>Step 0</b>. cd into an empty working directory.
 
  <b>Step 1</b>. Build Models
  
  Run the \ref builder
  	\code $ mpirun -np 8 -hostfile hostfile builder -l <install dir>/libs/libmxm_1d.so -U2000 -s100 \endcode
   
  Files made: 
  -	./conf_file 
  -	./\<hostname\>.\<node rank\>.\<device type\>.fpm
  	
  The default generated file ./conf_file is suitable if you only want to use cpu's.
  If you want to use some some other device (eg. gpu), 
  edit ./conf_file appropriately (and of course modify your routine to make use of this string). 
  An example of the format listed in \ref fupermod_conf_group

  \b Note: Do not mix up hostfile (a list of hostnames used by mpirun) 
  and ./conf_file (a list of hostnames, internal ranks, device type and other parameters for each MPI process, used by fupermod)

\n
  <b>Step 2</b>. Partition
  Run partitioner on head node (or locally) \ref partitioner
  \code $ partitioner -l <install dir>/libs/libmxm_1d.so -a3 -D256 \endcode
 
  Files made:
 	 - ./partition.dist
 
 <b>Step 3</b>. Convert 1D dist to 2D dist \ref convert_1dto2d_dist .
 \code $ 1dto2d_dist -k64 -N1024 \endcode
  Files made:
 	- ./partition.2ddist
 
 <b>Step 4</b>. Execute routine.
 \code $ mpirun -n 3 -hostfile machines <install dir>/mxm_2d \endcode
 

 \page software Software design
 Directories:
 * - /fupermod - \ref fupermod (core library)
 * - /tools - generic tools (applicable to most of routines)
 * - /routines - computational routines (applications to experiment with FPMs)
 * - /examples - standalone examples (not structured as routines)
 * - /cblas_wrappers - \ref fupermod_cblas (additional library, used in many routines and examples)
 * - /datapart_2d - 2D data partitioning based on FPMs (additional library, used in some routines)
 * - /tests - unit tests, prototypes
 * - /doxygen-doc - doxygen documentation in html and latex formats (generated by "make doxygen-doc")
  */

#endif /*DOXYMAIN_H_*/
