#include "fupermod_benchmark.h"
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <gsl/gsl_cdf.h>
#include <gsl/gsl_heapsort.h>
#include <gsl/gsl_statistics_double.h>

fupermod_precision fupermod_precision_defaults() {
  fupermod_precision precision;
  precision.reps_min = 3;
  precision.reps_max = 100;
  precision.cl = 0.95;
  precision.eps = 0.025;
  precision.time_max_rep = 60; // 1 minute
  // precision.zero_speed = 1e7; // 1e7 flops.
  precision.zero_speed = 0; // disabled! 	
  //precision.time_max_kill = 1200; // 20 minutes
  precision.time_max_kill = 0; // disabled!
  return precision;
}

double fupermod_ci(double cl, int reps, double* t) {
  return fabs(gsl_cdf_tdist_Pinv(cl, reps-1)) * gsl_stats_sd(t, 1, reps) / sqrt(reps);
}

int fupermod_point_read(FILE* stream, fupermod_point* point) {
  double tmp;
  return fscanf(stream, "%lld\t%le\t%le\t%d\t%le\t%le\n", &point->d, &point->t, &point->ci, &point->reps, &tmp, &tmp);
}

int fupermod_point_write(FILE* stream, fupermod_point* point, fupermod_complexity complexity) {
  double c = complexity(point->d, NULL);
  int res = fprintf(stream, "%lld\t%le\t%le\t%d\t%le\t%le\n", point->d, point->t, point->ci, point->reps, c, c / point->t);
  fflush(stream);
  return res;
}

int fupermod_point_node(fupermod_point* point, fupermod_point* point_node, MPI_Comm comm_intra, int root_intra) {
  fupermod_point res;
  MPI_Reduce(&point->d, &res.d, 1, MPI_LONG_LONG_INT, MPI_SUM, root_intra, comm_intra);
  MPI_Reduce(&point->t, &res.t, 1, MPI_DOUBLE, MPI_MAX, root_intra, comm_intra);
  MPI_Reduce(&point->ci, &res.ci, 1, MPI_DOUBLE, MPI_MAX, root_intra, comm_intra);
  MPI_Reduce(&point->reps, &res.reps, 1, MPI_INT, MPI_MAX, root_intra, comm_intra);
  int rank_intra;
  MPI_Comm_rank(comm_intra, &rank_intra);
  if (rank_intra == root_intra)
    *point_node = res;
  return 0;
}
