//todo: remove unnecessary
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gsl/gsl_heapsort.h>
#include "fupermod_conf.h"
#include "fupermod_debug.h"

void fupermod_conf_free(fupermod_process_conf conf) {
  free(conf.hostname);
  free(conf.bind);
  free(conf.device_type);
  free(conf.subopts);
//todo: remove the rest
  int i;
  for(i = 0; i < conf.argc; i++){
    free(conf.argv[i]);
  }
  free(conf.argv);
}

//local function
int match(const char *s, char p){
  int c = 1;
  while (*s != '\0') {
    if (strncmp(s++, &p, 1)) continue;
    c++;
  }
  return c;
}

//local function
int get_list_param(const char* input, int* argc, char*** argv){
  *argv = NULL;
  if(input==NULL){
    *argc = 0;
    return -1;
  }
  if(strlen(input) < 2) {
    *argc = 0;
    return -1;
  }
  char *line = strdup(input);
  int size = match(line, ' ');
  *argv = (char**)malloc(sizeof(char*) * size);
  char *pch = strtok (line," \t\n");
  int i = 0;
  while (pch != NULL)
  {
    if(strcmp(pch,"") != 0 ) {
      (*argv)[i] = strdup(pch);
      i++;
    }
    pch = strtok (NULL, " \t\n");
  }
  *argc = i;
  free(line); line = NULL;
  return 0;
}

//local function
//todo: is this really needed?
int* get_index(MPI_Comm comm, MPI_Comm comm_intra){
  int rank_intra = 0;
  int size;
  int* index;

  MPI_Comm_rank(comm_intra, &rank_intra);
  MPI_Comm_size(comm, &size);
  index = (int*)malloc(sizeof(int) * size);
  MPI_Allgather(&rank_intra, 1, MPI_INT, index, 1, MPI_INT, comm);
  return index;
}

fupermod_process_conf* fupermod_get_conf_all(char* filename, int* size){
  if (filename == NULL){
    fprintf(stderr, "Error filename null %s\n", __func__);
    MPI_Abort(MPI_COMM_WORLD, 200);
  }
  FILE* stream;
  stream = fopen(filename, "r");
  if (stream == NULL) {
    debug_print(0, (size_t) 0,(size_t) 0,
        "Try to open the configuration file %s\n", filename);
    perror("fopen");
    MPI_Abort(MPI_COMM_WORLD, 201);
  }
  char* line = NULL;
  size_t number = 0;
  int err;
  int n = 0;
  fupermod_process_conf* confs = NULL;
  while ((err = getline(&line, &number, stream)) != -1) {
    // skip comment headers
    if (line[0] == '#'|| line[0] == '\n')
      continue;
    confs = realloc(confs, sizeof(fupermod_process_conf) * (n + 1));
    confs[n].hostname = malloc(sizeof(char) * MPI_MAX_PROCESSOR_NAME);
    confs[n].bind = malloc(sizeof(char) * FUPERMOD_BIND_MAX_STRING);
    confs[n].device_type = malloc(sizeof(char) * FUPERMOD_DEVICE_MAX_STRING);
    confs[n].subopts = malloc(sizeof(char) * FUPERMOD_SUBOPTION_MAX_STRING);
    int pos = 0;
    int err = sscanf(line, "%s %d %s %s %s %n", confs[n].hostname, &confs[n].rank_intra, confs[n].bind, confs[n].device_type, confs[n].subopts, &pos);
    if (err < 4) {
      fprintf(stderr, "Error reading line%d: \"%s\" err:%d in %s\n", n, line, err, __func__);
      MPI_Abort(MPI_COMM_WORLD, 204);
    }
    if (err < 5) { //No subopts.
      confs[n].subopts[0] = '\0';
      confs[n].argc = 0;
      confs[n].argv = NULL;
    } else {
      get_list_param(&line[pos], &(confs[n].argc), &(confs[n].argv));
    }
    n++;
  }
  free(line);
  *size = n;
  return confs;
}

fupermod_process_conf* fupermod_get_conf_all_sorted(MPI_Comm comm, char* filename, int* size){
  //get all confs from file.
  fupermod_process_conf* confs = fupermod_get_conf_all(filename, size);

  int mpi_size;
  MPI_Comm_size(comm, &mpi_size);
  if (*size != mpi_size) {
    fprintf(stderr, "Error:conf file size != mpi_size %d != %d\n", *size, mpi_size);
    MPI_Abort(comm, 205);
  }
  
  int rank_intra;
  MPI_Comm comm_node;
  fupermod_comm_intra(comm, &comm_node);
  MPI_Comm_rank(comm_node, &rank_intra);
  MPI_Comm_free(&comm_node);

  char** names = fupermod_allgather_hostnames(comm);
  int* rank_intras = (int*)malloc(sizeof(int) * *size);
  MPI_Allgather(&rank_intra, 1, MPI_INT, rank_intras, 1, MPI_INT, comm);
  fupermod_process_conf* confs_sorted = (fupermod_process_conf*)malloc(sizeof(fupermod_process_conf) * *size);
  
  int i;
  for(i = 0; i < *size; i++){
    int j;
    for(j = 0; j < *size; j++){
      if (confs[j].rank_intra == rank_intras[i] && (strcmp(confs[j].hostname, names[i]) == 0)) {
        confs_sorted[i] = confs[j];
        break;
      }
    }
    if (j == *size) {
      fprintf(stderr, "Error: No configuration found for %s %d\n", names[i], rank_intras[i]);
      MPI_Abort(comm, 202);
    }
  }
  return confs_sorted;
}

fupermod_process_conf fupermod_get_conf(MPI_Comm comm, char* filename){
  //get all confs from file.
  int size;
  fupermod_process_conf* confs = fupermod_get_conf_all(filename, &size);

  //get hostname
  char name[MPI_MAX_PROCESSOR_NAME];
  int len;
  MPI_Get_processor_name(name, &len);

  int rank_intra;
  MPI_Comm comm_node;
  fupermod_comm_intra(comm, &comm_node);
  MPI_Comm_rank(comm_node, &rank_intra);
  MPI_Comm_free(&comm_node);
  fupermod_process_conf* conf = NULL;
  int i;
  for(i = 0; i < size; i++){
    if (confs[i].rank_intra == rank_intra && (strcmp(confs[i].hostname, name) == 0)) {
      conf = &(confs[i]);
      break;
    } else {
      fupermod_conf_free(confs[i]);
    }
  }
  if (conf == NULL) {
    debug_print(0, (size_t) 0, (size_t) 0,
        "No configuration found for %s %d\n", name, rank_intra );
    MPI_Abort(comm, 202);
  }
  //free(confs); confs = NULL;
  return *conf;
}

void fupermod_print_conf(MPI_Comm comm, int root, FILE* file, char* default_device_type, char* subopts){
  //TODO
  if(file == NULL) file = stdout;
  if(default_device_type == NULL) default_device_type = "cpu";

  MPI_Comm comm_node;
  fupermod_comm_intra(comm, &comm_node);

  char** names = fupermod_allgather_hostnames(comm);

  int* index = get_index( comm,  comm_node);
  MPI_Comm_free(&comm_node);
  int size;
  MPI_Comm_size(comm, &size);
  int i=0;
  int rank;
  MPI_Comm_rank(comm, &rank);
  if(rank == root){
    fprintf(file, "#hostname\trank_intra\tbind\tdevice_type\tsuboptions\n");
    for(i = 0; i < size; i++){
      fprintf(file, "%s\t%d\t%s\t%s\t%s\n", names[i], index[i], "na", default_device_type, subopts);
    }
  }
  free(names[0]);
  free(names);
  free(index);
  return;
}


int fupermod_comm_intra(MPI_Comm comm, MPI_Comm* comm_intra) {
  char name[MPI_MAX_PROCESSOR_NAME];
  int len;
  int size;
  char** names = fupermod_allgather_hostnames(comm);
  int color = -1;
  int i = 0;

  MPI_Get_processor_name(name, &len);
  MPI_Comm_size(comm, &size);
  while (i < size){
    if (strcmp(name, names[i]) == 0) {
      break;
    }
    i++;
  }
  color = i;
  free(names);
  // split the communicator
  int rank;
  MPI_Comm_rank(comm, &rank);
  return MPI_Comm_split(comm, color, rank, comm_intra);
}

int fupermod_comm_inter(MPI_Comm comm, MPI_Comm* comm_inter) {
  // collect processor names
  char name[MPI_MAX_PROCESSOR_NAME];
  int len;
  MPI_Get_processor_name(name, &len);
  int size;
  MPI_Comm_size(comm, &size);
  char* names = (char*)malloc(sizeof(char) * MPI_MAX_PROCESSOR_NAME * size);
  MPI_Allgather(name, MPI_MAX_PROCESSOR_NAME, MPI_CHAR, names, MPI_MAX_PROCESSOR_NAME, MPI_CHAR, comm);
  // sort processor names
  size_t* indices = (size_t*)malloc(sizeof(size_t) * size);
  gsl_heapsort_index(indices, names, size, sizeof(char) * MPI_MAX_PROCESSOR_NAME, (int(*)(const void*, const void*))strcmp);
  // colouring (0, key) for the first process on the processor, MPI_UNDEFINED elsewhere
  int color = MPI_UNDEFINED;
  int key = 0;
  int rank;
  MPI_Comm_rank(comm, &rank);
  int i = 0;
  // colour until the current process is found
  do
    if (i == 0 || strcmp(&names[MPI_MAX_PROCESSOR_NAME * indices[i - 1]], &names[MPI_MAX_PROCESSOR_NAME * indices[i]])) {
      if (indices[i] == (size_t) rank) {
        // the first process on the processor
        color = 0;
      } else {
        // the next processor
        key++;
      }
    }
  while (indices[i++] != (size_t) rank);
  free(names);
  free(indices);
  // split the communicator
  return MPI_Comm_split(comm, color, key, comm_inter);
}

char** fupermod_gather_hostnames(int root, MPI_Comm comm) {
  char** friendly_names = NULL;
  char* hostnames = NULL;
  char hostname[MPI_MAX_PROCESSOR_NAME];
  int len;
  MPI_Get_processor_name(hostname, &len);
  int rank;
  MPI_Comm_rank(comm, &rank);
  int size;
  if (rank == root){
    MPI_Comm_size(comm, &size);
    hostnames = (char*)malloc(sizeof(char) * MPI_MAX_PROCESSOR_NAME * size);
  }
  MPI_Gather(hostname, MPI_MAX_PROCESSOR_NAME, MPI_CHAR, hostnames, MPI_MAX_PROCESSOR_NAME, MPI_CHAR, root, comm);
  if (rank == root){
    int i;
    for(i = 0; i < size;i++) friendly_names[i] = &hostnames[MPI_MAX_PROCESSOR_NAME * i];
  }
  return friendly_names;
}

char** fupermod_allgather_hostnames(MPI_Comm comm) {
  char name[MPI_MAX_PROCESSOR_NAME];
  int len;
  int size;
  int i;
  char** friendly_names;
  char*  names;

  MPI_Get_processor_name(name, &len);
  MPI_Comm_size(comm, &size);
  friendly_names = malloc(sizeof(char*) * size);
  names          = malloc(sizeof(char) * MPI_MAX_PROCESSOR_NAME * size);

  MPI_Allgather(name, MPI_MAX_PROCESSOR_NAME, MPI_CHAR, names, MPI_MAX_PROCESSOR_NAME, MPI_CHAR, comm);

  for( i = 0; i < size;i++) friendly_names[i] = &names[MPI_MAX_PROCESSOR_NAME * i];
  return friendly_names;
}

// ============
// Obsolete code, to delete soon
/*int fupermod_host_device_name(fupermod_device_info* info, char* host_device_name) {
  int pos = strcspn(info->hostname, ".");
  sprintf(host_device_name, "%.*s.%d.%s", pos, info->hostname, info->rank_intra, fupermod_device_enum_to_string(info->device_type));
  return 0;
  }

  int fupermod_read_device_file(char* nodetype_path, char* hostname, int* node_size, char** device_strings){
  char filename[1024];
  int pos = strcspn(hostname, ".");
  sprintf(filename, "%s.%.*s.dev", nodetype_path, pos, hostname);
 *node_size = -1;

 FILE* stream = NULL;
 if ((stream = fopen(filename, "r")) != NULL) {
 if (fscanf(stream, "#Devices %d\n", node_size)!= 1 ||
 *node_size < 0 || *node_size > 1e4) {
 fprintf(stderr, "Error reading from %s\n"
 "\tFirst line wasn't as expected: %d, should be like:\n"
 "\t #Devices 4\n", filename, *node_size);
 exit(14);
 }
 *device_strings = (char*)malloc(sizeof(char) * FUPERMOD_MAX_DEVICE_STRING * *node_size);
 int i;
 int rank;
 int rc;
 for (i = 0; i < *node_size; i++) {
 rc = fscanf(stream, "%d\t%s\n", &rank, &(*device_strings)[i * FUPERMOD_MAX_DEVICE_STRING]);
 if (rc != 2){
 fprintf(stderr, "Error reading from %s\n"
 "failed to read line %d\n", filename, i);
 exit(15);
 }
 if (rank != i) {
 fprintf(stderr, "Error reading from %s\n"
 "number in first colum not as expected\n"
 "%d != %d\n", filename, rank, i);
 exit(16);
 }
 }
 fclose(stream);
 } else {
 fprintf(stderr, "Warning: Could not open node discription file: %s\n"
 "\tsetting node_size to zero\n", filename);
//couldn't open file:
 *node_size = 0;
 }
 return 0;
 }

 int fupermod_read_device_info_all(char* nodetype_path, int machinefile_size, char* machinefile_hostnames, int* num_nodes_, int** num_devices_, char** unique_hostnames_, ... device_types_){
// sort processor names and pull out unique hosts.
size_t* indices = (size_t*)malloc(sizeof(size_t) * machinefile_size);
gsl_heapsort_index(indices, machinefile_hostnames, machinefile_size, sizeof(char) * MPI_MAX_PROCESSOR_NAME, (int(*)(const void*, const void*))strcmp);
char* unique_hostnames = (char*)malloc(sizeof(char) * MPI_MAX_PROCESSOR_NAME * machinefile_size);
int num_nodes = 0;
int i;
for (i = 0; i < machinefile_size; i++) {
if (i == 0 || strcmp(&machinefile_hostnames[MPI_MAX_PROCESSOR_NAME * indices[i - 1]], &machinefile_hostnames[MPI_MAX_PROCESSOR_NAME * indices[i]])) {
strcpy(&unique_hostnames[MPI_MAX_PROCESSOR_NAME * num_nodes], &machinefile_hostnames[MPI_MAX_PROCESSOR_NAME * indices[i]]);
num_nodes++;
}
}
free(indices); indices = NULL;
int* num_devices = (int*)malloc(sizeof(int) * num_nodes);
fupermod_device_type** device_types = (fupermod_device_type**)malloc(sizeof(fupermod_device_type*) * num_nodes);
//read device files
int total_devices = 0;
for (i = 0; i < num_nodes; i++) {
char* device_strings = NULL;
fupermod_read_device_file(nodetype_path, &unique_hostnames[MPI_MAX_PROCESSOR_NAME * i], &num_devices[i], &device_strings);
if (num_devices[i] < 1) { //if no node-type file, make it work anyway.
fprintf(stderr, "Warning: Using default device: %d\n", FUPERMOD_DEFAULT_DEVICE);
device_types[i] =(fupermod_device_type*)malloc(sizeof(fupermod_device_type) * 1);
device_types[i][0] = FUPERMOD_DEFAULT_DEVICE;
num_devices[i] = 1;
} else {
  device_types[i] =(fupermod_device_type*)malloc(sizeof(fupermod_device_type) * num_devices[i]);
  int j;
  for (j = 0; j < num_devices[i]; j++) {
    device_types[i][j] = fupermod_device_string_to_enum(&device_strings[j * FUPERMOD_MAX_DEVICE_STRING]);
  }
}
total_devices+= num_devices[i];
free(device_strings); device_strings = NULL;
}
*num_nodes_ = num_nodes;
*num_devices_ = num_devices;
*unique_hostnames_ = unique_hostnames;
*device_types_ = device_types;
return total_devices;
}

fupermod_device_info* fupermod_device_read_info(MPI_Comm comm, char* nodetype_path){
  //my hostname
  char hostname[MPI_MAX_PROCESSOR_NAME];
  int len;
  MPI_Get_processor_name(hostname, &len);
  //my rank_intra
  MPI_Comm comm_intra;
  fupermod_comm_intra(comm, &comm_intra);
  int rank_intra;
  MPI_Comm_rank(comm_intra, &rank_intra);
  //my device type
  return fupermod_device_read_info_file(nodetype_path, hostname, rank_intra);
}
void fupermod_device_info_free(fupermod_device_info* info){
  free(info->hostname);
  free(info);
}

fupermod_device_info* fupermod_device_read_info_file(char* nodetype_path, char* hostname, int rank_intra){
  fupermod_device_info* info = (fupermod_device_info*)malloc(sizeof(fupermod_device_info));
  info->hostname = (char*)malloc(sizeof(char) * MPI_MAX_PROCESSOR_NAME);
  strcpy(info->hostname, hostname);
  info->rank_intra = rank_intra;

  int node_size = -1;
  char* device_strings = NULL;
  fupermod_read_device_file(nodetype_path, hostname, &node_size, &device_strings);

  if (rank_intra >= node_size) {
    fprintf(stderr, "Error: rank_intra >= nodesize, %d >= %d\n"
        "\tChange number of hosts in mpirun or fix nodetype file\n", rank_intra, node_size);
    exit(17);
  }
  if (node_size < 1) {
    fprintf(stderr, "Warning: Using default device: %d\n", FUPERMOD_DEFAULT_DEVICE);
    info->device_type = FUPERMOD_DEFAULT_DEVICE;
  } else {
    info->device_type = fupermod_device_string_to_enum(&device_strings[rank_intra * FUPERMOD_MAX_DEVICE_STRING]);
  }
  free(device_strings); device_strings = NULL;
  return info;
}
*/

/*fupermod_device_type fupermod_device_string_to_enum(char* device_string){
  fupermod_device_type type;
  if (strcmp(device_string, "cpu") == 0)
  type = cpu;
  else if (strcmp(device_string, "gpu") == 0)
  type = gpu;
  else {
  fprintf(stderr, "Error: fupermod_device string:%s not recognised. %s\n", device_string, __FUNCTION__);
  exit(12);
  }
  return type;
  }

  const char* fupermod_device_enum_to_string(fupermod_device_type type){
  switch(type) {
  case cpu: 
  return "cpu";
  break;
  case gpu:
  return "gpu";
  break;
  default:
  fprintf(stderr, "Error: fupermod_device enum:%d not recognised. %s\n", type, __FUNCTION__);
  exit(13);
  }
  return NULL;
  }

  }*/
