#include "config.h"
#include "fupermod_data.h"
#include <stdlib.h>
#include <memory.h>
#include <stdio.h>
#include <time.h>

fupermod_data* fupermod_data_alloc() {
  fupermod_data* data = (fupermod_data*)malloc(sizeof(fupermod_data));
  data->count = 0;
  data->points = NULL;
  return data;
}

void fupermod_data_free(fupermod_data* data) {
  if (data != NULL) {
    free(data->points);
    free(data);
  }
}

int fupermod_data_read(FILE* stream, fupermod_data* data) {
  while (!feof(stream)) {
    fupermod_point point;
    if (fupermod_point_read(stream, &point) > 0) {
      data->points = (fupermod_point*)realloc(data->points, sizeof(fupermod_point) * (data->count + 1));
      data->points[data->count] = point;
      data->count++;
    } else {
      //skip over header lines
      int c;
      while ((c = getc(stream)) != '\n' && c != EOF);
    }
  }
  return 0;
}

int fupermod_data_write_head(FILE* stream, fupermod_library* library) {
  char hostname[MPI_MAX_PROCESSOR_NAME];
  int len;
  MPI_Get_processor_name(hostname, &len);
  time_t now = time(NULL);
  struct tm *tm_now = localtime(&now);
  char datetime[25];
  strftime(datetime, sizeof(datetime), "%d-%m-%y %H:%M", tm_now);
  fprintf(stream, "#Kernel:%s %s\n", library->path, library->conf->subopts);
  fprintf(stream, "#Machine:%s at:%s\n", hostname, datetime);
  fprintf(stream, "#problem\ttime\tconfidence\treps\tcomplexity\tspeed\n");
  return 0;
}

int fupermod_data_write(FILE* stream, fupermod_data* data, fupermod_library* library) {
  fupermod_data_write_head(stream, library);
  int i;
  for (i = 0; i < data->count; i++)
    fupermod_point_write(stream, &data->points[i], ((fupermod_kernel*)(library->kernel))->complexity);
  return 0;
}

int fupermod_data_add(fupermod_data* data, fupermod_point point) {
  if (point.d > 0 && point.t > 0) {
    int found = 0;
    int i;
    for (i = 0; i < data->count; i++) {
      if (point.d == data->points[i].d) {
        // Harmonic mean
        data->points[i].t = 2 / (1.0 / data->points[i].t + 1.0 / point.t);
        data->points[i].reps++;
        found = 1;
        break;
      }
    }
    if (!found) {
      data->points = (fupermod_point*)realloc(data->points, sizeof(fupermod_point) * (data->count + 1));
      if (data->count == 0 || point.d > data->points[data->count - 1].d) {
        data->points[data->count] = point;
      } else if (point.d < data->points[0].d) {
        memmove(&(data->points[1]), data->points, sizeof(fupermod_point) * data->count);
        data->points[0] = point;
      } else {
        int i;
        for (i = 0; i < data->count - 1; i++)
          if (point.d > data->points[i].d && point.d < data->points[i + 1].d) {
            memmove(&(data->points[i + 2]), &(data->points[i + 1]), sizeof(fupermod_point) * (data->count - i - 1));
            data->points[i + 1] = point;
            break;
          }
      }
      data->count++;
    }
  }
  return 0;
}

char* fupermod_data_filename(const char* path, fupermod_process_conf conf) {
  char* filename = NULL;
  asprintf(&filename, "%s%s.%d.%s.fpm", path, conf.hostname, conf.rank_intra, conf.device_type);
  return filename;
}
