#ifndef FUPERMOD_DATA_H_
#define FUPERMOD_DATA_H_

#include "fupermod_benchmark.h"

#ifdef __cplusplus
extern "C" {
#endif
  /*!
   * \defgroup fupermod_data_group Data points
   * API for input/output the data obtained from benchmarks. The main data structure is \ref fupermod_data,
   * which consists of experimental points \ref fupermod_point.
   * \{
   */

  /*!
   * Data points sorted by problem size.
   * Problem sizes must be positive.
   * Life cycle: created before model, destroyed after.
   * To add a data point, use \ref fupermod_data_add.
   */
  typedef struct fupermod_data {
    /*! Number of points */
    int count;
    /*! Points */
    fupermod_point* points;
  } fupermod_data;

  /*!
   * Allocates the data
   */
  fupermod_data* fupermod_data_alloc();

  /*! Frees the data */
  void fupermod_data_free(fupermod_data* data);

  /*! Reads data from a file */
  int fupermod_data_read(FILE* stream, fupermod_data* data);

  /*! Writes header info to a file */
  int fupermod_data_write_head(FILE* stream, fupermod_library* library);

  /*! Writes data to a file */
  int fupermod_data_write(FILE* stream, fupermod_data* data, fupermod_library* library);

  /*!
   * Adds a data point
   * \param data data
   * \param point point
   * \return index or -1
   */
  int fupermod_data_add(fupermod_data* data, fupermod_point point);


  /*! Returns pointer to allocated memory containing the name of a data file in the form: "path.hostname.rank_intra.device_type.fpm". 
   * Expect the tool to already have populated conf with appropriate info.
   * User must free the allocated memory manually
   * \param path initial path
   * \param conf configuration to generate the rest of the file name
   */
  char* fupermod_data_filename(const char* path, fupermod_process_conf conf);

  /*!
   * \}
   */
#ifdef __cplusplus
}
#endif

#endif /*FUPERMOD_DATA_H_*/
