#include "fupermod_debug.h"
#include "fupermod_dynamic.h"
#include <stdlib.h>
#include <float.h>
#include <memory.h>
#include <math.h>

int fupermod_partitioner_iterate(fupermod_dynamic* partitioner, MPI_Comm comm, int root, fupermod_precision precision, fupermod_benchmark* benchmark, double eps) {
  // benchmark
  long long int d;
  fupermod_dist_scatter(partitioner->dist, &d, root, comm);
  fupermod_point point;
  //TODO need to fix this if want to use dypartitioner for many MPI processes per node.
  MPI_Comm comm_benchmark_barrier = MPI_COMM_SELF;
  int rc_execute = benchmark->execute(benchmark, comm_benchmark_barrier, d, precision, &point);
  if (rc_execute == -1) {
    fprintf(stderr, "Execution of kernel failed, returned -1.  %s\n", __func__);
    MPI_Abort(comm, 71);
  }
  int status = 0;

  // Gather and update models
  int rank;
  MPI_Comm_rank(comm, &rank);
  //fprintf(stderr, "P%d: Result of benchmark->execute: %d %s\n",rank, rc_execute, __func__);
  if (rank == root) {
    fupermod_point* points = (fupermod_point*)malloc(sizeof(fupermod_point) * partitioner->size);
    MPI_Gather(&point, sizeof(fupermod_point), MPI_CHAR, points, sizeof(fupermod_point), MPI_CHAR, root, comm);

    int* rc_executes = (int*)malloc(sizeof(int) * partitioner->size);
    MPI_Gather(&rc_execute, 1, MPI_INT, rc_executes, 1, MPI_INT, root, comm);
    int i;
    for (i = 0; i < partitioner->size; i++) {
      if (rc_executes[i] == -2) {
        status = -2;
        fprintf(stderr, "Warning: Execution of kernel returned -2 on P%d, but continuing.  %s\n", i, __func__);
        break;
      }
    }
    free(rc_executes);

    for (i = 0; i < partitioner->size; i++) {
      //fupermod_data_add(partitioner->models[i]->data, points[i]);
      partitioner->models[i]->update(partitioner->models[i], points[i]);
    }
    // measure disbalance:
    if (status == 0) 
      status = fupermod_dist_test_times(points, partitioner->size, eps);
    // status is now 1 if should finish because error is small
    free(points);
    // partition
    fupermod_dist* new_dist = fupermod_dist_alloc(partitioner->dist->D, partitioner->size);
    partitioner->partition(new_dist, partitioner->size, partitioner->models);

    // did the dist change? 
    if (status == 0) {
      status = 1;
      for (i = 0; i < partitioner->size; i++) {
        if (partitioner->dist->parts[i].d != new_dist->parts[i].d){
          status = 0;
          break;
        }
      }
    }
    fupermod_dist_free(partitioner->dist);
    partitioner->dist = new_dist;
  } else {
    MPI_Gather(&point, sizeof(fupermod_point), MPI_CHAR, NULL, 0, MPI_DATATYPE_NULL, root, comm);
    MPI_Gather(&rc_execute, 1, MPI_INT, NULL, 0, MPI_INT, root, comm);
  }
  MPI_Bcast(&status, 1, MPI_INT, root, comm);
  return status;
  // If status is: 
  // 0: benchmark worked and should continue iterating, 
  // less than 0 iteration should not be incremented. 
  // Greater than 0: distribution has converged.
}

int fupermod_partitioner_iterate_one(fupermod_dynamic* partitioner, MPI_Comm comm, int root, fupermod_precision precision, fupermod_benchmark* benchmark, double eps) {
  NOT_USED(eps);
  // benchmark
  long long int d = 1;
  fupermod_point point;
  int rc_execute = benchmark->execute(benchmark, comm, d, precision, &point);
  if (rc_execute == -1) {
    fprintf(stderr, "Execution of kernel failed, returned -1.  %s\n", __func__);
    MPI_Abort(comm, 71);
  }

  // Gather and update models
  int rank;
  MPI_Comm_rank(comm, &rank);
  fprintf(stderr, "P%d: Result of benchmark->execute: %d %s\n",rank, rc_execute, __func__);
  if (rank == root) {
    fupermod_point* points = (fupermod_point*)malloc(sizeof(fupermod_point) * partitioner->size);
    MPI_Gather(&point, sizeof(fupermod_point), MPI_CHAR, points, sizeof(fupermod_point), MPI_CHAR, root, comm);

    int i;
    for (i = 0; i < partitioner->size; i++) {
      partitioner->models[i]->update(partitioner->models[i], points[i]);
    }
    free(points);
  } else {
    MPI_Gather(&point, sizeof(fupermod_point), MPI_CHAR, NULL, 0, MPI_DATATYPE_NULL, root, comm);
  }
  return 0;
}
