#include "fupermod_kernel.h"
#include <dlfcn.h>
#include <stdio.h>
#include <string.h>

int fupermod_library_load(MPI_Comm comm, char* path, fupermod_process_conf* conf, fupermod_library* library) {
  library->path = path;
  library->conf = conf;
  char name[MPI_MAX_PROCESSOR_NAME];
  if (comm == MPI_COMM_NULL) 
    strcpy(name, "localhost");
  else {
    int len;
    MPI_Get_processor_name(name, &len);
  }
  library->handle = dlopen(path, RTLD_LAZY);
  if (library->handle == NULL) {
    fprintf(stderr, "%s: %s\n", name, dlerror());
    return -1;
  }
 
  library->kernel = dlsym(library->handle, FUPERMOD_KERNEL_INSTANCE);
  if (library->kernel == NULL) {
    fprintf(stderr, "%s: %s\n", name, dlerror());
    return -1;
  }
  int err = library->kernel->load(comm, conf);
  if (err) {
    fprintf(stderr, "%s: kernel load failed with the error code %d\n", name, err);
    return -1;
  }
  return 0;
}

int fupermod_library_unload(MPI_Comm comm, fupermod_library* library) {
  char name[MPI_MAX_PROCESSOR_NAME];
  if (comm == MPI_COMM_NULL) 
    strcpy(name, "localhost");
  else {
    int len;
    MPI_Get_processor_name(name, &len);
  }
  int err = library->kernel->unload(comm);
  if (err) {
    fprintf(stderr, "%s: kernel unload failed with the error code %d\n", name, err);
    return -1;
  }
  err = dlclose(library->handle);
  if (err) {
    fprintf(stderr, "%s: %s\n", name, dlerror());
    return -1;
  }
  return 0;
}
