#include "config.h"
#include "fupermod_model.h"
#include <stdio.h>
#include <unistd.h>

double fupermod_model_speed(fupermod_model* model, long long int d) {
  return model->complexity(d, NULL) / model->t(model, d);
}

void fupermod_print_model(FILE* stream, fupermod_library* library, fupermod_process_conf conf, fupermod_model* model, long long int D) {
  char hostname[MPI_MAX_PROCESSOR_NAME];
  gethostname(hostname, MPI_MAX_PROCESSOR_NAME);
  time_t now = time(NULL);
  struct tm *tm_now = localtime(&now);
  char datetime[25];
  strftime(datetime, sizeof(datetime), "%d-%m-%y %H:%M", tm_now);
  fprintf(stream, "#Model of %s %d %s %s %s\n", conf.hostname, conf.rank_intra, conf.bind, conf.device_type, conf.subopts);
  fprintf(stream, "#Model used by:%s at:%s for problem size %lld\n", hostname, datetime, D);
  fprintf(stream, "#Kernel:%s\n", library->path);
  fprintf(stream, "#problem\ttime\tnull\tnull\tcomplexity\tspeed\ttime derivative\n");

  D *= 2; // uncomment to see full range.
  double step, max_steps = (D > 1000)? 1000 : D;
  for (step = 0; step < max_steps; step++) {
    long long int d = step * D / max_steps;
    double c = model->complexity(d, NULL);
    double t = model->t(model, d);
    fprintf(stream, "%lld\t%le\t.\t.\t%le\t%le\t%le\n", d, t, c, c / t, model->dt(model, d));
  }
}

void fupermod_print_models_data(FILE* stream, const int size, fupermod_model** models) {
  int l = 0;
  int cont = 1;
  int i;
  fprintf(stream, "#");
  for (i = 0; i < size; i++) {
    fprintf(stream, "d[%d]\tt[%d]\t", i, i);
  }
  fprintf(stream, "\n");
  while (cont){
    cont = 0;
    for (i = 0; i < size; i++) {
      if (l < models[i]->data->count) {
        cont = 1;
        fprintf(stream, "%lld\t%le\t", models[i]->data->points[l].d, models[i]->data->points[l].t);
      } else {
        fprintf(stream, ".\t.\t");
      }
    }
    fprintf(stream, "\n");
    l++;
  }
}

char* fupermod_model_filename(const char* path, fupermod_process_conf conf) {
  char* filename = NULL;
  asprintf(&filename, "%s%s.%d.%s.model", path, conf.hostname, conf.rank_intra, conf.device_type);
  return filename;
}
