#include "fupermod_model.h"
#include "fupermod_debug.h"
#include <stdlib.h>

/*! The constant model. Uses the Akima spline interpolation of real observations */
typedef struct fupermod_model_constant {
  /*! Base */
  fupermod_model base;
  /*! A data point whose time approximation will be used as the value of the constant model */
  fupermod_point last_point;
} fupermod_model_constant;


double fupermod_model_constant_t(struct fupermod_model* model, double x) {
  fupermod_model_constant* constant = (fupermod_model_constant*)model;
  if (model->data->count < 1){
    fprintf(stderr, "Error: looking for a point from constant model when no points have been added yet. %s\n", __func__);
    return -1;
  }
  if (constant->last_point.d > 0){
    return constant->last_point.t * model->complexity(x, NULL) / model->complexity(constant->last_point.d, NULL);
  } else {
    fprintf(stderr, "Info: Using first point in data file. %s\n", __func__);
    return model->data->points[0].t * model->complexity(x, NULL) / model->complexity(model->data->points[0].d, NULL);
  }
}

double fupermod_model_constant_dt(struct fupermod_model* model, double x) {
  NOT_USED(model);
  NOT_USED(x);
  return 0;
}

int fupermod_model_constant_update(fupermod_model* model, fupermod_point point) {
  if (point.d > 0){
    fupermod_data_add(model->data, point);
    ((fupermod_model_constant*)model)->last_point = point;
  }
  return 0;
}

fupermod_model* fupermod_model_constant_alloc(fupermod_data* data, fupermod_complexity complexity, long long int D, void* params, int smooth) {
  NOT_USED(params);
  NOT_USED(smooth);
  fupermod_model_constant* constant = (fupermod_model_constant*)malloc(sizeof(fupermod_model_constant));
  constant->last_point = (fupermod_point){0, 0, 0, 0};
  fupermod_model* model = (fupermod_model*)constant;
  model->data = data;
  model->complexity = complexity;
  model->D = D;
  model->t = fupermod_model_constant_t;
  model->dt = fupermod_model_constant_dt;
  model->update = fupermod_model_constant_update;
  fupermod_model_constant_update(model, (fupermod_point){0, 0, 0, 0});
  return model;
}

void fupermod_model_constant_free(fupermod_model* model) {
  free(model);
}
