#!/usr/bin/env python

"""
This script builds Functional Performance models (FPMs) of an application.
This script iterates through different problem sizes.
For each problem size: 
    It runs the application through the extra profiler 
    http://www.bsc.es/computer-sciences/extrae 
    This generate a trace file.
    The script trace2FPMpoint.py parses this tracefile and for each MPI process adds one point to the FPM for this problem size.

"""
  
__appname__ = "Builder_extrae"
__author__  = "Dave Clarke & HCL"
__version__ = "0.1.0"
__license__ = "GNU GPL 3.0 or later"
__doc__ = ""
  
import logging
log = logging.getLogger(__name__)
import subprocess
import trace2FPMpoint

import sys
import shutil
#sys.path.append(sys.path[0] + "/../routines/gadget/")
from gadget_execute_extrae import *

DEFAULT_MIN = 1000
DEFAULT_MAX = 2000
DEFAULT_D   = 4000 
  
global opts
print(sys.path[0])
#execute_script = sys.path[0] + "/../routines/mxm/mxm_execute_extrae.py"
#execute_script = sys.path[0] + "/../routines/gadget/gadget_execute_extrae.py"

def main():
    # Initialise
    routine = Gadget(opts.D, opts.MPI_size)
    
    # Execute 
    if (opts.steps < 2 or opts.upperN <= opts.lowerN):
        increment = opts.upperN
    else:
        increment = int((opts.upperN - opts.lowerN) / (opts.steps - 1.0))
    if (increment < 1):
        increment = 1;
    N = opts.lowerN
    while(N <= opts.upperN):

        print(N)
        routine.execute_benchmark(N) 

        execute_str = """mpi2prv  -f TRACE.mpits -o gadget_sic1.prv"""
        print("Exec:" + execute_str)
        subprocess.check_call(execute_str, shell=True)

        trace2FPMpoint.trace2fpm_point("gadget_sic1", MPI_size)

        shutil.rmtree('set-0', ignore_errors=True)

        N += increment

    # Finilise
    routine.finalise()





if __name__ == '__main__':
        from optparse import OptionParser
        parser = OptionParser(version="%%prog v%s" % __version__,
                        usage="%prog [options] <argument> ...",
                        description=__doc__.replace('\r\n', '\n').split('\n--snip--\n')[0])
        parser.add_option('-v', '--verbose', action="count", dest="verbose",
                default=2, help="Increase the verbosity. Use twice for extra effect")
        parser.add_option('-q', '--quiet', action="count", dest="quiet",
                default=0, help="Decrease the verbosity. Use twice for extra effect")
        parser.add_option('-p', '--np', dest="MPI_size",
                default=2, help="Number of mpi processes", type="int")
        parser.add_option('-U', dest="upperN",
                default=DEFAULT_MAX, help="Upper problem size [default: %default]", type="int")
        parser.add_option('-L', dest="lowerN",
                default=DEFAULT_MIN, help="Lower problem size [default: %default]", type="int")
        parser.add_option('-s', dest="steps",
                default=3, help="number of steps in the model [default: %default]", type="int")
        parser.add_option('-D', dest="D",
                default=DEFAULT_D, help="Overall problem size D [default: %default]", type="int")

  
        # Allow pre-formatted descriptions
        parser.formatter.format_description = lambda description: description
  
        opts, args  = parser.parse_args()
  
        # Set up clean logging to stderr
        log_levels = [logging.CRITICAL, logging.ERROR, logging.WARNING,
                                    logging.INFO, logging.DEBUG]
        opts.verbose = min(opts.verbose - opts.quiet, len(log_levels) - 1)
        opts.verbose = max(opts.verbose, 0)
        MPI_size = opts.MPI_size
        logging.basicConfig(level=log_levels[opts.verbose],
                                                format='%(levelname)s: %(message)s')
        main()
