#!/usr/bin/env python

"""

"""
 
__appname__ = "Gadget class for Builder_extrae"
__author__  = "Dave Clarke & HCL"
__version__ = "0.1.0"
__license__ = "GNU GPL 3.0 or later"
__doc__ = ""
 
import sys
import subprocess
import random

#sys.path.append(sys.path[0] + "/../../fupermod")
#print(sys.path[0] + "/../../fupermod")
import fupermod

N = 0
k = 4

MIN_N = 1

class Gadget:
    overall_D = 0
    MPI_size = 0
    hostnames = []


    def __init__(self, D, size):
        Gadget.overall_D = D
        Gadget.MPI_size= size
        execute_str = """gdl -e "setupsic, '""" + str(D) + """'" """
        print("Exec:" + execute_str)
        subprocess.check_call(execute_str, shell=True)
        Gadget.hostnames = fupermod.parse_conf_file()        
        if (len(Gadget.hostnames) != Gadget.MPI_size):
            raise IndexError("Size not the same as in conf file. %d != %d", len(Gadget.hostnames), Gadget.MPI_size)


    def compute_partition_even(self):
        print("** Doing even partitioning **\n")
        dist = [0]*Gadget.MPI_size
        for i in range(Gadget.MPI_size):
            dist[i] = Gadget.overall_D / Gadget.MPI_size + (1 if i < Gadget.overall_D % Gadget.MPI_size else  0)
        return dist

    def compute_partition_alternate(self, N):
        print("** Doing alternating partitioning based on %d **\n" %N)
        size = Gadget.MPI_size
        D = Gadget.overall_D
        dist = [0]*size
        import math
        lower = int(math.ceil(size / 2))
        upper = size - lower

        lowerN = N*lower
        lowerN = min((D - upper * MIN_N), lowerN)
        upperN = D - lowerN
        
        altN = upperN / upper
        dist[::2] = [N]*lower
        dist[1::2] = [altN]*upper

        print("Before fix:")
        print(dist, sum(dist))
        i = 0
        remainder = D - sum(dist)  
        while remainder > 0:
            dist[i] += 1
            remainder -= 1
            i += 1
            if (i >= Gadget.MPI_size): i = 1

        i = 0
        while remainder < 0:
            if dist[i] > MIN_N:
                dist[i] -= 1
                remainder += 1
            i += 1
            if (i >= Gadget.MPI_size): i = 1

        print("After fix:")
        print(dist, sum(dist))
        if sum(dist) != Gadget.overall_D:
            raise IndexError("Array does not sum up. %d != %d", sum(dist), Gadget.overall_D)

        return dist


    def compute_partition_random(self, N):
        print("** Doing random partitioning based on %d**\n" %N)
        size = Gadget.MPI_size
        D = Gadget.overall_D
        dist = [0]*size

        dist[0] = N # set p0

        #distribute remainder
        remainder = (D  - N)

        sum_array = N
        for i in range(1, size):
            dist[i] = random.randint(MIN_N, remainder)
            sum_array += dist[i]
        remainder = D - sum_array  
        i = 1
        while remainder > 0:
            dist[i] += 1
            remainder -= 1
            i += 1
            if (i >= size): i = 1

        i = 1
        while remainder < 0:
            if dist[i] > MIN_N:
                dist[i] -= 1
                remainder += 1
            i += 1
            if (i >= size): i = 1

        if sum(dist) != D:
            raise IndexError("Array does not sum up. %d != %d", sum(dist), D)

        return dist

    def compute_partition_random_select(self, N):
        print("** Doing random partitioning based on %d**\n" %N)
        size = Gadget.MPI_size
        D = Gadget.overall_D
        dist = [(D  - N)/(size - 1)]*size
        selected = random.randint(0, size-1)
        dist[selected] = N # set p0

        #distribute remainder
        print("Before fix:")
        print(dist, sum(dist))

        remainder = D - sum(dist)  
        i = 0
        while remainder > 0:
            dist[i] += 1
            remainder -= 1
            i += 1
            if (i >= size): i = 0

        i = 0
        while remainder < 0:
            if dist[i] > MIN_N:
                dist[i] -= 1
                remainder += 1
            i += 1
            if (i >= size): i = 0

        print("After fix:")
        print(dist, sum(dist))

        if sum(dist) != D:
            raise IndexError("Array does not sum up. %d != %d", sum(dist), D)

        return dist


    def create_partition_file(self, dist):
        file = open("partition.dist", "w")
        file.write("#size %d\n" % Gadget.MPI_size)
        file.write("#D %d\n" % Gadget.overall_D)
        file.write("#host	rank_intra	d	        time	i	x	  complexity	       speed\n")
        for i,host in enumerate(Gadget.hostnames):
            file.write("%s %d %d 0 0 0 %d 0\n" %(host[0], host[1], dist[i], dist[i]))
        file.close

    def execute_benchmark(self, N):
        if (N + MIN_N * (Gadget.MPI_size - 1)) > Gadget.overall_D:
            print("Warning: N + min > overall_D. N:%d  D:%d\n" % (N, Gadget.overall_D))
            N = Gadget.overall_D - MIN_N * (Gadget.MPI_size - 1)
            print("Warning: Adjusted to N:%d\n" % (N))
        if N < MIN_N:
            print("Warning: N < MIN_N. N:%d\n" % (N))
            N = MIN_N
            print("Warning: Adjusted to N:%d\n" % (N))

            #return 0

        #dist = self.compute_partition(N)
        #dist = self.compute_partition_even()
        #dist = self.compute_partition_alternate(N)
        dist = self.compute_partition_random_select(N)
        self.create_partition_file(dist)

        execute_str = """mpirun -np """ + str(Gadget.MPI_size) + """ --machinefile machinefile trace.sh Gadget2 SIC_gen.param """
        print("Exec:" + execute_str)
        subprocess.check_call(execute_str, shell=True)
        sys.stdout.flush()

    def finalise(self):
        print('finalise function')
