#!/usr/bin/env python

# Extract a single point from an extrae trace file.

__appname__ = "trace2FPMpoint"
__author__  = "Dave Clarke & HCL"
__version__ = "0.1.5"
__license__ = "GNU GPL 3.0 or later"
__doc__ = ""

import logging
log = logging.getLogger(__name__)

import sys
import re
def printf(format, *args):
        sys.stdout.write(format % args)

def print_file(file, format, *args):
        file.write(format % args)

#globals:

def main():
    #Options
    from optparse import OptionParser
    parser = OptionParser(version="%%prog v%s" % __version__,
                    usage="%prog [options] <argument> ...",
                    description=__doc__.replace('\r\n', '\n').split('\n--snip--\n')[0])
    parser.add_option('-v', '--verbose', action="count", dest="verbose",
            default=2, help="Increase the verbosity. Use twice for extra effect")
    parser.add_option('-q', '--quiet', action="count", dest="quiet",
            default=0, help="Decrease the verbosity. Use twice for extra effect")
    parser.add_option("-f", "--file", dest="filename",
                                help="Basename of .prv .row files to read from. Required", metavar="FILE")
    parser.add_option("-e", "--event", dest="event_of_interest",
                                help="Extrae event to measeure [default: %default].", metavar="EVENT_NUM", default=95000001)
    parser.add_option('-p', '--np', dest="MPI_size",
            default=1, help="Number of mpi processes", type="int")

    # Allow pre-formatted descriptions
    parser.formatter.format_description = lambda description: description

    opts, args  = parser.parse_args()

    # Set up clean logging to stderr
    log_levels = [logging.CRITICAL, logging.ERROR, logging.WARNING,
                                logging.INFO, logging.DEBUG]
    opts.verbose = min(opts.verbose - opts.quiet, len(log_levels) - 1)
    opts.verbose = max(opts.verbose, 0)
    logging.basicConfig(level=log_levels[opts.verbose],
                                            format='%(levelname)s: %(message)s')


    event_of_interest = int(opts.event_of_interest)
    logging.info('Event of interest: %d', event_of_interest)
    if not opts.filename:
        parser.error('Filename not given')

    #End options
    trace2fpm_point(opts.filename, opts.MPI_size, event_of_interest)

def trace2fpm_point(filename, MPI_size, event_of_interest = 95000001):
    logging.debug('filename: %s', filename)
    events_to_remove = 50000001, 50000002, 50000003
    N_event = 95000500 #problem size in run.
    complexity_event = 95000501
    overall_size_event = 95000502
    events_to_record = [N_event, complexity_event, overall_size_event]
    proc_names = parse_row_file(filename + ".row", MPI_size)
    rank_intras, problem_sizes = parse_dist_file(proc_names)
    # problem_sizes and Ns should be the same - but from different sources so probably not problem_size is requested by partitioner, Ns is actual.
    comp_time,iters,recorded_events = parse_prv_file(MPI_size, filename + ".prv", event_of_interest, events_to_remove, events_to_record)
    #todo:check comp_time not all zero.
    print_point_to_files(proc_names, rank_intras, MPI_size, comp_time, iters, problem_sizes, recorded_events)

def print_point_to_files(proc_names, rank_intras, MPI_size, comp_time, iters, problem_sizes, recorded_events):
    for i in range(0, MPI_size):
        if comp_time[i] != 0:
            filename = proc_names[i] + '.' + str(rank_intras[i]) + '.cpu.fpme'
            f = open(filename, 'a')
            print_file(f, '%d\t%e\t0\t%d\t%e\t%e\t%d\n', recorded_events[0][i], comp_time[i], iters[i], recorded_events[1][i], recorded_events[1][i]/comp_time[i], recorded_events[2][i])
            f.close
            logging.info('Sucessfully written to %s', filename)
        else:
            logging.warning('Not adding a point for process %s %d because time recorded is zero.', proc_names[i], rank_intras[i])

def parse_row_file(filename, MPI_size):
    f = open(filename, 'r')

    #first read num of mpi processes => size
    l = f.readline()
    p = re.compile('LEVEL CPU SIZE (\d+)')
    m = p.match(l)
    if not m:
        logging.critical('Failed to match first line in %s', filename)
        logging.critical(l)
        size = 1
    else:
        size = int(m.group(1))
    if MPI_size != size:
        raise IndexError("Size does not match %d %d", MPI_size, size)

    #2. Get mpi process names...  to format: host.rank_intra
    proc_names = ['']*size
    proc_ex_nums = [0]*size

    p = re.compile(r'(\d+)\.([A-Za-z\d\.-]{1,63})')
    for i in range(0, size):
        l = f.readline()
        m = p.match(l)
        if not m:
            logging.critical('Failed to match hostname on line:')
            logging.critical(l)
        else:
            proc_names[i] = m.group(2)  
            proc_ex_nums[i] = int(m.group(1)) #may not be needed.
    f.close


    return proc_names

def parse_dist_file(proc_names):
    index = []
    rank_intras = [-1]*len(proc_names)
    problem_sizes = [-1]*len(proc_names)
    for line in open("partition.dist"):
        li=line.strip()
        if not li.startswith("#"):
            dat = line.split()
            print(dat[0])
            #match = [item for item in proc_names if item == dat[0]]
            for i, val in enumerate(proc_names):
                if i not in index:
                    if (val == dat[0]):
                        index.append(i)
                        rank_intras[i] = int(dat[1])
                        problem_sizes[i] = int(dat[2])
                        print("i:", i, "val:", val, dat[1])
                        break

    if -1 in rank_intras:
        logging.critical("Didn't match all rank intras")
        logging.critical(rank_intras)
        raise Missmatch_error("Didn't match all rank intras")
    if -1 in problem_sizes:
        raise Missmatch_error("Didn't match all problem sizes %s",  problem_sizes)
    print(index)
    print(proc_names)
    print(problem_sizes)
    return rank_intras, problem_sizes


def parse_prv_file(MPI_size, filename, event_of_interest, events_to_remove, events_to_record):

    current_start_time = [-1]*MPI_size
    inside_loop = [0]*MPI_size
    remove_time = [0]*MPI_size
    inside_remove = [0]*MPI_size
    current_remove_time = [-1]*MPI_size
    current_remove_event = [-1]*MPI_size
    computational_time = [0]*MPI_size #this is final result.
    remove_time_total = [0]*MPI_size
    iterations = [0]*MPI_size
    recorded_events = [[-1]*MPI_size, [-1]*MPI_size, [-1]*MPI_size]

    with open(filename) as f:
        for fullline in f:
            sline = fullline.split(':')
            if len(sline) > 7:
                try:
                    linetype = int(sline[0])
                except ValueError:
                    logging.warning("Could not parse line:")
                    logging.warning(fullline)
                else:
                    if  linetype == 2: #If it is 1, line is about state not event.
                        try:
                            rank = int(sline[1]) - 1 #extrae ranks start at 1. Here ranks start at 0.
                            time = int(sline[5])
                            event = int(sline[6])
                            value = int(sline[7])
                        except  ValueError:
                            logging.warning("Could not convert value on line:")
                            logging.warning(fullline)
                        else: #have line can work with.
                            if rank >= MPI_size:
                                logging.error("rank out of range: %s. MPI_size: %s", rank, MPI_size)
                                raise IndexError("Rank out of range.")
                            # 1. Is this event of interest?
                            if event == event_of_interest:
                                #found event of interest.
                                logging.debug("Found event of interest!")
                                logging.debug(fullline)
                                if inside_loop[rank]: #2nd matching point found.
                                    if inside_remove[rank]:
                                        logging.critical("End of loop, but still inside remove loop. Don't trust answer.")
                                        logging.critical(fullline)
                                    this_time_gross = time - current_start_time[rank]
                                    this_time_nett = this_time_gross - remove_time[rank]
                                    logging.debug('P: %d gross: %d, \tremove: %d nett: %d', rank, this_time_gross, remove_time[rank], this_time_nett)
                                    computational_time[rank] += this_time_nett
                                    remove_time_total[rank] += remove_time[rank] #for information only.
                                    iterations[rank] += 1 #work out the avg time for one loop.
                                    current_start_time[rank] = time
                                    remove_time[rank] = 0
                                    if value == 0: #leaveing loop
                                        inside_loop[rank] = 0
                                elif value > 0: #found first point - entering loop.
                                    inside_loop[rank] = 1
                                    current_start_time[rank] = time
                                    #2. Is this an event to remove?
                            elif event in events_to_remove and inside_loop[rank]:
                                logging.debug("Removing: %s", fullline)
                                if inside_remove[rank]: #inside_remove
                                    if current_remove_event[rank] != event:
                                        logging.critical('Remove event %d != current remove event %d.'
                                                'Need to write a more complex parser for this!', current_remove_event[rank], event)
                                        if time < current_remove_time[rank]:
                                            logging.critical('Time difference negative! %d %d', time, current_remove_time[rank])
                                    logging.debug('P%d: Removing:%d - %d = %d', rank, time, current_remove_time[rank], time - current_remove_time[rank])
                                    remove_time[rank] +=  time - current_remove_time[rank]
                                    if value > 0: #going into another remove segment, so reset.
                                        current_remove_time[rank] = time
                                        current_remove_event[rank] = event
                                    else:
                                        inside_remove[rank] = 0
                                        current_remove_time[rank] = -1
                                        current_remove_event[rank] = -1
                                else: #not inside_remove
                                    current_remove_event[rank] = event
                                    current_remove_time[rank] = time
                                    inside_remove[rank] = 1
                            else:
                                for idx, revent in enumerate(events_to_record):
                                    if event == revent:
                                        if recorded_events[idx][rank] != -1 and recorded_events[idx][rank] != value:
                                            logging.warning("P%d: Event to record(%d) redefined to new value. Was: %d. Now: %d.", rank, idx, recorded_events[idx][rank], value)
                                        recorded_events[idx][rank] = value
                                        logging.info("P%d: N: %d", rank, recorded_events[idx][rank])
    f.closed

    for rank in range(MPI_size):
        logging.info("P%d: Comp time: %f Removed time: %f total time: %f\n", rank, computational_time[rank]/1e9, remove_time_total[rank]/1e9, (computational_time[rank] + remove_time_total[rank])/1e9)

    #convert nano-seconds to seconds.
    #and get avg time of one iteration.
    for rank in range(MPI_size):
        if iterations[rank] > 0:
            computational_time[rank] = computational_time[rank]/1e9/iterations[rank]

    return(computational_time, iterations, recorded_events)




if __name__ == '__main__':
    main()
