#include "mxm_2d.h"
#include <stdlib.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <memory.h>
#ifdef EXTRAE_ENABLED
#include <extrae_user_events.h>
#endif

#define VERBOSE_MATRIX 0
#define dx(i) idist->d[i].x
#define dy(i) idist->d[i].y
#define dw(i) idist->d[i].w
#define dh(i) idist->d[i].h
#define di(k) idist->d[k].i
#define dj(k) idist->d[k].j
#define PRINT_TIMES 0

int MPI_MSG_TAG_FOR_A = 0xf1;
int MPI_MSG_TAG_FOR_B = 0xf2;

fupermod_dgemm_workspace* fupermod_dgemm_alloc(MPI_Comm comm, FILE* stream, fupermod_process_conf* confs, fupermod_gemm* gemm) {
  fupermod_dgemm_workspace* w = (fupermod_dgemm_workspace*)malloc(sizeof(fupermod_dgemm_workspace));
  w->gemm = gemm;

  int rank;
  MPI_Comm_rank(comm, &rank);
  int size; 
  MPI_Comm_size(comm, &size);

  // compute idist.
  //double* areas = (double*)malloc(sizeof(double) * size);
  //MPI_Allgather(&area, 1, MPI_LONG_LONG_INT, areas, 1, MPI_DOUBLE, comm);
  //w->idist = mm_col_dist(size, N, kb, areas);
  //TODO check the size after the read
  w->idist = mm_col_dist_read(stream, size, confs);
  if (w->idist == NULL)
    return NULL;
  int kb = w->idist->kb;

  // print total volume of communication.
  {
    int tvc = 0;
    if (rank == 0){
      int i;
      for (i = 0; i < size; i++) {
        tvc += w->idist->d[i].w;
        tvc += w->idist->d[i].h;
      }
      fprintf(stdout, "tvc: %d\n", tvc);
    }
  }
  //free(areas); areas = NULL;

  int myw = w->idist->d[rank].w;
  int myh = w->idist->d[rank].h;
  // working space:
  w->WA = fupermod_malloc(sizeof(fupermod_float) * myh * kb * kb, confs+rank);
  w->WB = fupermod_malloc(sizeof(fupermod_float) * myw * kb * kb, confs+rank);

  if (w->WA == NULL || w->WB == NULL) {
    fprintf(stderr, "P%d Failed to allocate matrix\n"
        "P%d myw = %d, myh = %d, kb = %d\n"
        ,rank, rank, myw, myh, kb);
    fupermod_dgemm_free(w, confs+rank);
    return NULL;
  }

  return w;
}

void fupermod_dgemm_free(fupermod_dgemm_workspace* w, fupermod_process_conf* conf){
  fupermod_free(w->WA, conf);
  fupermod_free(w->WB, conf);

  mm_col_i_dist_free(w->idist);
  free(w); w = NULL;
}

// todo: should be someting like: 
/* int fupermod_dgemm(enum CBLAS_ORDER Order, enum CBLAS_TRANSPOSE TransA, enum CBLAS_TRANSPOSE TransB, const int M, const int N, const int K,
   double alpha, double *A, const int lda, double *B, const int ldb, double beta, double *C, const int ldc, MPI_Comm comm, mm_col_i_distribution idist) */
int fupermod_dgemm(fupermod_float *A, fupermod_float *B, fupermod_float *C, MPI_Comm comm, fupermod_dgemm_workspace* w) {


  int kb = w->idist->kb;
  fupermod_float* WA = w->WA;
  fupermod_float* WB = w->WB;

  int rank;
  MPI_Comm_rank(comm, &rank);
  int size; 
  MPI_Comm_size(comm, &size);
  mm_col_i_distrib* idist = w->idist;
  int myx, myy, myw, myh, myi, myj;
  myx = idist->d[rank].x;
  myy = idist->d[rank].y;
  myw = idist->d[rank].w;
  myh = idist->d[rank].h;
  myi = idist->d[rank].i;
  myj = idist->d[rank].j;

  int lda, ldb, ldc; 
  lda = ldb = ldc = myw * kb;

  int rc;
  MPI_Comm col_comm; 
  rc = MPI_Comm_split(comm, myj, myi, &col_comm); //color: processes in 'th column, key: i'th process in column
  if (rc != MPI_SUCCESS) {
    fprintf(stderr, "Error splitting groups\n");
    return -1;
  }

  // For non blocking MPI_Irecv
  MPI_Request* requestsS = (MPI_Request*)malloc(sizeof(MPI_Request)*(size-1) * 2);
  MPI_Request* requestsR = (MPI_Request*)malloc(sizeof(MPI_Request)*(size-1) * 2);
  MPI_Status* statusesS=(MPI_Status*)malloc(sizeof(MPI_Status)*(size-1));
  MPI_Status* statusesR=(MPI_Status*)malloc(sizeof(MPI_Status)*(size-1));
  if (requestsS == NULL || requestsR == NULL || statusesS == NULL || statusesR == NULL){
    fprintf(stderr, "P%d Failed to allocate MPI objects\n", rank);
    return -1;
  }
  fupermod_float* tmp = (fupermod_float*) malloc(sizeof(fupermod_float) * 10 * size);
  if ( tmp == NULL ){
    fprintf(stderr, "P%d Failed to allocate space for overhead\n", rank);
    return -1;
  }
  free(tmp);

  //non-blocking index 
  int recv_count, send_count = 0;

  // progress report
  struct timeval start, inter1, inter2;
  gettimeofday(&start, NULL);
  gettimeofday(&inter2, NULL);
  double total_time_dgemm = 0;

  FILE* time_file = NULL;
  if(PRINT_TIMES){
    char filename[256];
    sprintf(filename, "times.mxmcol.%d.dat", rank);
    time_file = fopen(filename, "w");
    char hostname[MPI_MAX_PROCESSOR_NAME];
    int len;
    MPI_Get_processor_name(hostname, &len);
    fprintf(time_file, "# host: %s, myw:%e myh:%le myi:%d myj:%d N:%le\n", hostname, (double)myw, (double)myh, myi, myj, (double)w->idist->N);
  }


#ifdef EXTRAE_ENABLED
	extrae_type_t extrae_t1 = 95000001;
	extrae_type_t extrae_t2 = 95000500;
	unsigned nvalues1 = 0;
	Extrae_define_event_type(&extrae_t1 , "DGEMM loop", &nvalues1, NULL, NULL);
	Extrae_define_event_type(&extrae_t2 , "Complexity 1 loop", &nvalues1, NULL, NULL);
  Extrae_event(extrae_t2, 2 * myh * kb * myw * kb * kb);
#endif


  // To speed up experiments: export MXM_2D_STOP=10
  int stop_at = w->idist->N;
  if (getenv("MXM_2D_STOP") != NULL) {
    int stop_early = atoi(getenv("MXM_2D_STOP"));
    if (stop_early > 0 && stop_early < w->idist->N) {
      stop_at = stop_early; //modification to speed up experiments.
      fprintf(stderr, "P%d: Warning only doing %d iterations.\n", rank, stop_at);
    }
  }
  // If zero work to do, skip loop:
  if (myw < 1 || myh < 1)
    stop_at = -1;

  // Main k loop - moving pivrow & pivcol	
  int k;
  for (k = 0; k < stop_at; k++) {
#ifdef EXTRAE_ENABLED
    Extrae_event(extrae_t1, k+1);
#endif
    // progress report
    if (rank == 0 && k > 0) {
      gettimeofday(&inter1, NULL);
      if ( inter1.tv_sec - inter2.tv_sec > 60) {
        double time = ((double)(inter1.tv_sec - start.tv_sec) + (inter1.tv_usec - start.tv_usec) * 1.0e-6);
        double expected = time / (double)k * w->idist->N;
        fprintf(stderr, "loop %d of %d, expected time: %le\n", k, w->idist->N, expected);
        gettimeofday(&inter2, NULL);
      }
    }

    fupermod_float* A_pt = WA;
    fupermod_float* B_pt = WB;
    lda = kb;
    ldb = myw * kb;
    recv_count = 0;
    send_count = 0;

    // Find j index of procs holding pivcol
    int pivcol_j = -1;
    int i;
    for (i = 0; i < size; i++) {
      if (k >= dx(i) && k < dx(i)+dw(i)) {
        pivcol_j = dj(i);
        break;
      }
    }
    // Am i pivcol?
    int pivcol = 0;
    if (myj == pivcol_j) {
      pivcol = 1;
      A_pt = &A[(k - myx) * kb]; // point to local A.
      lda = myw * kb; // lda changes.
    }
    int pivrow_proc = -1;
    for (i = 0; i < size; i++) {
      if ((dj(i) == myj) && k >= dy(i) && k < dy(i)+dh(i)) {
        pivrow_proc = i;
        break;
      }
    }
    if (pivcol_j < 0) {
      fprintf(stderr, "Error, Problem with pivcol\n");
    }
    if (pivrow_proc < 0) {
      fprintf(stderr, "P%d: Error, Problem with pivrow_proc\n", rank);
      return -1;
    }

    // Send/recv	pivcol of A
    // test intersection with each proc.
    for (i = 0; i < size; i++) {
      //MPI_Barrier(comm);
      if ((pivcol || dj(i) == pivcol_j) && i != rank && // either i'm pivcol or proc i is
          (myy < dy(i) + dh(i)) && (myy + myh > dy(i))) { // there is an overlap

        //start send/recv from
        int start = dy(i) - myy;
        if (start < 0) {
          start = 0;
        }
        int length;
        if (dy(i) + dh(i) > myy + myh) {
          length = myh - start;
        } else {
          length = dy(i) + dh(i) - myy - start;
        }
        if (pivcol) { // send
          MPI_Datatype	fupermod_col;
          MPI_Type_vector(length * kb, kb, myw * kb, FUPERMOD_MPI_FLOAT, &fupermod_col);
          MPI_Type_commit(&fupermod_col);

          rc = MPI_Isend(&A[(start * kb) * (myw * kb) + (k - myx) * kb], 1, fupermod_col, i, MPI_MSG_TAG_FOR_A, comm, &requestsS[send_count++]);
          if (rc != MPI_SUCCESS) {
            fprintf(stderr, "Error Sending A\n");
            return -1;
          }
          MPI_Type_free(&fupermod_col);
        } else { // recv
          rc = MPI_Irecv(&WA[start * kb * kb], length * kb * kb, FUPERMOD_MPI_FLOAT, i, MPI_MSG_TAG_FOR_A, comm, &requestsR[recv_count++]);
          if (rc != MPI_SUCCESS) {
            fprintf(stderr, "Error Recving A\n");
            return -1;
          }
        }
      }
    }

    if (pivrow_proc == rank) {
      rc = MPI_Bcast(&B[(k - myy) * kb * kb * myw], myw * kb * kb, FUPERMOD_MPI_FLOAT, myi, col_comm);
      B_pt = &B[(k - myy) * kb * kb * myw]; // Set for local dgemm
    } else {
      rc = MPI_Bcast(WB, myw * kb * kb, FUPERMOD_MPI_FLOAT, di(pivrow_proc), col_comm);
    }
    if (rc != MPI_SUCCESS) {
      fprintf(stderr, "Error Bcat'n B\n");
      return -1;
    }

    // Wait for all recv's to complete.
    MPI_Waitall(recv_count, requestsR, statusesR);
    // Optimisation: wait for all sends to complete.
    MPI_Waitall(send_count, requestsS, statusesS);

    if (VERBOSE_MATRIX && rank == 0) {
      int j;
      printf("P%d A(%d,%d):\n", rank, myh * kb, kb);
      for (i = 0; i < myh * kb; i++) {
        for (j = 0; j < kb; j++) {
          printf("  %3.1f", A_pt[i * lda + j]);
        }
        printf("\n");
      }
      printf("P%d B(%d,%d):\n", rank, myh * kb, kb);
      for (i = 0; i < kb; i++) {
        for (j = 0; j < ldb; j++) {
          printf("  %3.1f", B_pt[i * ldb + j]);
        }
        printf("\n");
      }
    }
    struct timeval start_dgemm, end_dgemm;
    gettimeofday(&start_dgemm, NULL);
    fupermod_gemm_execute(w->gemm, CblasRowMajor, CblasNoTrans, CblasNoTrans, myh * kb, myw * kb, kb, 1, A_pt, lda, B_pt, ldb, 1, C, myw * kb);
    gettimeofday(&end_dgemm, NULL);
    double time = ((double)(end_dgemm.tv_sec - start_dgemm.tv_sec) + (end_dgemm.tv_usec - start_dgemm.tv_usec) * 1.0e-6);
    total_time_dgemm += time;
    if (PRINT_TIMES)
      fprintf(time_file, "%le\n", time);

    if (VERBOSE_MATRIX) {
      if (rank == 0) {
        printf("P%d C(%d,%d):\n", rank, myh * kb, myw * kb);
        int j;
        for (i = 0; i < myw * kb; i++) {
          for (j = 0; j < myh * kb; j++) {
            printf("  %3.3f", C[i * myw * kb + j]);
          }
          printf("\n");
        }
      }
    }
  }// end k loop
#ifdef EXTRAE_ENABLED
  Extrae_event(extrae_t1, 0);
#endif

  //finish all communication before exiting
  //MPI_Waitall(send_count, requestsS, statusesS);
  MPI_Barrier(comm);

  struct timeval end;
  gettimeofday(&end, NULL);
  double total_time = ((double)(end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec) * 1.0e-6);
  double total_comm = total_time - total_time_dgemm;
  if(PRINT_TIMES){
    fprintf(time_file, "#P%d: total: %le dgemm: %le comm: %le ratio dgemm/comm : %4.2lf\n", rank, total_time, total_time_dgemm, total_comm, total_time_dgemm / total_comm);
    fclose(time_file);
  }
  fprintf(stdout, "#P%d: total: %le dgemm: %le comm: %le ratio dgemm/comm : %4.2lf\n", rank, total_time, total_time_dgemm, total_comm, total_time_dgemm / total_comm);
#ifdef GREEN_COMPUTING
  FILE* green_file = NULL;
  char green_fname[256];
  sprintf(green_fname, "green.mxmcol.%d.%d.dat", rank, w->idist->N);
  green_file = fopen(green_fname, "w");
  if(green_file !=NULL) {
    char hostname[MPI_MAX_PROCESSOR_NAME];
    int len;
    MPI_Get_processor_name(hostname, &len);
    fprintf(green_file, "# host: %s, myw:%e myh:%le myi:%d myj:%d N:%le\n", hostname, (double)myw, (double)myh, myi, myj, (double)w->idist->N);
    double total_time = ((double)(end.tv_sec - start.tv_sec) + (end.tv_usec - start.tv_usec) * 1.0e-6);
    double total_comm = total_time - total_time_dgemm;
    fprintf(green_file, "#P%d: total: %le dgemm: %le comm: %le ratio dgemm/comm : %4.2lf start : %lf end : %lf\n", rank, total_time, total_time_dgemm, total_comm, total_time_dgemm / total_comm, ((double)(start.tv_sec) + start.tv_usec * 1.0e-6), ((double)(end.tv_sec) + end.tv_usec * 1.0e-6));
    fclose(green_file);
  } else {
    fprintf(stderr, "Unable to open file for writing green computing data.\n");
  }
#endif
  free(requestsS);
  free(statusesS);
  free(requestsR);
  free(statusesR);
  return 0;
}

