/*!
 * \page mxm_col Matrix multiplication col
 * Implements the model-based BR algorithm
 * todo: to rewrite
 * \see \subpage convert1dto2d_mxm_dist
 */
#include "config.h"
#include "fupermod/fupermod_partition.h"
#include "mxm_2d.h"
#include "cblas_wrappers/fupermod_cblas.h"
#ifdef HAVE_LIBCUBLAS
#include "cblas_wrappers/cuda_auxiliary.hpp"
#endif
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#ifdef HAVE_LIBMKL_CORE
#include <mkl_service.h>
#endif

#define VERBOSE_MATRIX 0

/*! Fill matrix with increasing numbers - For testing purpose. */
void fill_matrix_testing(int rank, fupermod_dgemm_workspace* workspace, fupermod_float *A, fupermod_float *B, fupermod_float *C);
/*! Fill matrix with constant numbers. */
void fill_matrix_constants(int rank, fupermod_dgemm_workspace* workspace, fupermod_float *A, fupermod_float *B, fupermod_float *C);
/*! Test: compare solution with a serial mxm. */
void compare_solution(int rank, fupermod_dgemm_workspace* workspac, fupermod_float *C, MPI_Comm comm);

int main(int argc, char** argv) {
  MPI_Init(&argc, &argv);
  MPI_Comm comm = MPI_COMM_WORLD;

  // exit routine
  int exit = 0; 
  // Test final result.
  int test_result = 0;
  //partition file path
  char dist_path[256] = "./partition.2ddist";
  char conf_file[256] = "./conf_file";

  int rank;
  MPI_Comm_rank(comm, &rank);
  int size;
  MPI_Comm_size(comm, &size);
  int root = 0;

  //if (!rank) getc(stdin); MPI_Barrier(MPI_COMM_WORLD);
  //if (rank == root) {
  int opt;
  while ((opt = getopt(argc, argv, "htp:m:")) != -1)
    switch(opt) {
      case 'h':
        if (rank == root) {
          fprintf(stderr,
              "routine\n"
              "Usage: [options]\n"
              "  -h  help\n"
              "  -t  Test final result (default none)\n"
              "  -p S  Path to partition file (default %s)\n"
              "  -m S  Path to conf file (default %s)\n",
              dist_path, conf_file);
        }
        exit = 1;
        break;
      case 't':
        test_result = 1;
        break;
      case 'p':
        strcpy(dist_path, optarg);
        break;
      case 'm':
        strcpy(conf_file, optarg);
        break;
      default:
        fprintf(stderr, "Unknown option %s\n", optarg);
        break;
    }
  //}

  MPI_Bcast(&exit, 1, MPI_INT, root, comm);
  if (exit) {
    MPI_Finalize();
    return 0;
  }

  int conf_file_size;
  fupermod_process_conf* confs = fupermod_get_conf_all_sorted(comm, conf_file, &conf_file_size); // where do you free?
  
  #ifdef HAVE_LIBMKL_CORE
  mkl_set_dynamic(0);
  int threads =atoi(confs[rank].subopts);
  mkl_set_num_threads(threads);
  #endif

  //initialize
  fupermod_gemm* gemm = fupermod_gemm_alloc(confs+rank);

  // read from dist file and allocate working space
  fupermod_dgemm_workspace* workspace = NULL;
  FILE* stream = NULL;
  if ((stream = fopen(dist_path, "r")) != NULL) {
    workspace = fupermod_dgemm_alloc(comm, stream, confs, gemm);
    if (workspace == NULL)
      MPI_Abort(comm, 51); 
    fclose(stream);
  } else {
    fprintf(stderr, "Error: Could not open dist file: %s\n", dist_path);
    MPI_Abort(comm, 50);
  }

  if (rank == 0)
    mm_col_print(workspace->idist);

  int myw = workspace->idist->d[rank].w;
  int myh = workspace->idist->d[rank].h;

  // initialise A, B, C, WA, WB
  int kb = workspace->idist->kb;
  size_t mem_size = sizeof(fupermod_float) * (myw * kb * myh * kb);

  fupermod_float *A = NULL;
  fupermod_float *B = NULL;
  fupermod_float *C = NULL;
  A = (fupermod_float*)malloc(mem_size);
  B = (fupermod_float*)malloc(mem_size);
  C = (fupermod_float*)fupermod_malloc(mem_size, confs+rank);

  // Did it work?
  if (A == NULL || B == NULL || C == NULL) {
    fprintf(stderr, "P%d Failed to allocate matrix\n"
        "P%d myw = %d, myh = %d, kb = %d\n"
        "P%d A:%p B:%p C:%p\n"
        "p%d Exiting\n",rank, rank, myw, myh, kb, rank, A, B, C, rank);
    free(A);  A = NULL; 
    free(B);  B = NULL;
    fupermod_free(C, confs+rank); C = NULL;
    fupermod_gemm_free(gemm);
    fupermod_dgemm_free(workspace, confs+rank);
    MPI_Finalize();
    return -1;
  }
  // Fill A, B. Zero C
  if (test_result)
    fill_matrix_testing(rank, workspace, A, B, C);
  else
    fill_matrix_constants(rank, workspace, A, B, C);

  if (!strcmp((confs+rank)->device_type, "gpu")) {
    #ifdef HAVE_LIBCUBLAS
    cublasxgemm_params* params = (cublasxgemm_params*)gemm->params;
    params->num_iteration = workspace->idist->N;
    params->workspace_WA = workspace->WA;
    params->workspace_WB = workspace->WB;
    #else
    fprintf(stderr, "%s : %s: Error no cuda library\n", __FILE__, __func__);
    #endif
  }

  // fupermod_dgemm
  fupermod_dgemm(A, B, C, comm, workspace);

  if (test_result)
    compare_solution(rank, workspace, C, comm);

  free(A);  A = NULL;
  free(B);  B = NULL;
  fupermod_free(C, confs+rank); C = NULL;
  fupermod_dgemm_free(workspace, confs+rank);
  fupermod_gemm_free(gemm);
  MPI_Finalize();
  return 0;
}

void fill_matrix_testing(int rank, fupermod_dgemm_workspace* workspace, fupermod_float *A, fupermod_float *B, fupermod_float *C){
  int kb = workspace->idist->kb;
  if (rank == 0) 
    mm_col_print(workspace->idist);

  int kw = kb * workspace->idist->d[rank].w;
  int kh = kb * workspace->idist->d[rank].h;
  int kx = kb * workspace->idist->d[rank].x;
  int ky = kb * workspace->idist->d[rank].y;
  int kN = kb * workspace->idist->N;

  int i, j;
  for (i = 0; i < kh; i++) {
    for (j = 0; j < kw; j++) {
      A[i * kw + j] = (i + ky) * kN + (j + kx);
      B[i * kw + j] = (i + ky) * kN + (j + kx) + 5;
    }
  }

  for (i = 0; i < kh* kw; i++) {
    C[i] = 0.0;
  }
}

void fill_matrix_constants(int rank, fupermod_dgemm_workspace* workspace, fupermod_float *A, fupermod_float *B, fupermod_float *C){
  int kb = workspace->idist->kb;
  int myw = workspace->idist->d[rank].w;
  int myh = workspace->idist->d[rank].h;

  int i;
  for (i = 0; i < myh * kb * myw * kb; i++) {
    A[i] = 1.1;
    B[i] = 2.5;
    C[i] = 0.0;
  }
}

void compare_solution(int rank, fupermod_dgemm_workspace* workspace, fupermod_float *C, MPI_Comm comm){
  int kb = workspace->idist->kb;
  int kw = kb * workspace->idist->d[rank].w;
  int kh = kb * workspace->idist->d[rank].h;
  int kx = kb * workspace->idist->d[rank].x;
  int ky = kb * workspace->idist->d[rank].y;
  int kN = kb * workspace->idist->N;

  fupermod_float* At = (fupermod_float*) malloc(sizeof(fupermod_float) * (kN * kN));
  fupermod_float* Bt = (fupermod_float*) malloc(sizeof(fupermod_float) * (kN * kN));
  fupermod_float* Ct = (fupermod_float*) fupermod_malloc(sizeof(fupermod_float) * (kN * kN), workspace->gemm->conf);

  int i, j;
  for (i = 0; i < kN * kN; i++) {
    At[i] = i;
    Bt[i] = i + 5;
  }
  for (i = 0; i < kN * kN; i++) {
    Ct[i] = 0.0;
  }

  if (!strcmp(workspace->gemm->conf->device_type, "gpu")) {
    #ifdef HAVE_LIBCUBLAS
    cublasxgemm_params* params = (cublasxgemm_params*)workspace->gemm->params;
    params->idx_outer = 0; // so that cuda_gemm will alloc device memory dA/B/C for computing Ct
    params->num_iteration = 1; // just one iteration in compare_solution
    params->algorithm = -1; // switch to basic algorithm
    #else
    fprintf(stderr, "%s : %s: Error no cuda library\n", __FILE__, __func__);
    #endif
  }
  fupermod_gemm_execute(workspace->gemm, CblasRowMajor, CblasNoTrans, CblasNoTrans, kN, kN, kN, 1, At, kN, Bt, kN, 0, Ct, kN);

  double diff = 0;
  for (i = 0; i < kh; i++) {
    for (j = 0; j < kw; j++) {
      double max = fmax(fabs(Ct[(i + ky) * kN + (j + kx)]), fabs(C[i * kw + j]));
      if (max > 0)
        diff += fabs(Ct[(i + ky) * kN + (j + kx)] - C[i * kw + j]) / max;
       // printf("%f %f\n",Ct[(i + ky) * kN + (j + kx)], C[i * kw + j]);
    }
  }
  double sum;
  MPI_Reduce(&diff, &sum, 1, MPI_DOUBLE, MPI_SUM, 0, comm);
  if (rank == 0) {
    if (sum < 1E-6)
      fprintf(stderr, "Matrix multiplication was successful\n");
    else 
      fprintf(stderr, "***\nError: Matrix multiplication failed, diff = %le\n***\n", sum);
  }

  free(At); At = NULL;
  free(Bt); Bt = NULL;
  fupermod_free(Ct, workspace->gemm->conf); Ct = NULL;
}
