#include "mxm_car.h"
#include "fupermod/fupermod_debug.h"
#include "fupermod/fupermod_benchmark.h"
#include "cblas_wrappers/fupermod_cblas.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
/*
#include <omp.h>
omp_set_num_threads(6);
*/
#define VERBOSE_MATRIX 0
void verbose_matrix(double* ptr, int height, int width);

#define dx(i) idist->d[i].x
#define dy(i) idist->d[i].y
#define dw(i) idist->d[i].w
#define dh(i) idist->d[i].h
#define di(k) idist->d[k].i
#define dj(k) idist->d[k].j

int MPI_MSG_TAG_FOR_A = 0xf1;
int MPI_MSG_TAG_FOR_B = 0xf2;

//fupermod_dgemm_workspace* fupermod_dgemm_alloc(MPI_Comm comm, int N, int kb, long long int area) {
fupermod_dgemm_workspace* fupermod_dgemm_alloc_car(MPI_Comm comm, int N, int kb, long long int area, fupermod_process_conf* confs, char* distfile) {
  NOT_USED(N);
  NOT_USED(area);
  fupermod_dgemm_workspace* w;
  w = (fupermod_dgemm_workspace*)malloc(sizeof(fupermod_dgemm_workspace));
  w->kb = kb;
  w->idist = NULL;

  int root = 0;
  int rank;
  MPI_Comm_rank(comm, &rank);
  int size; 
  MPI_Comm_size(comm, &size);

  if (confs == NULL && distfile == NULL) {	
    return NULL;
  } else {
    if (rank == root) {
      FILE* stream = NULL;
      if ((stream = fopen(distfile, "r")) != NULL) {
        if (!(w->idist = mm_col_dist_read(stream, size, confs))) {
          fprintf(stderr, "Error: read partition.dist.car: %s\n", distfile);
          fclose(stream); 
          fupermod_dgemm_free_car(w);
          return NULL;
        }
      } else {
        fprintf(stderr, "Error: open partition.dist.car: %s\n", distfile);
        fclose(stream); 
        fupermod_dgemm_free_car(w);		
        return NULL;
      }
      fclose(stream);
    }
    int rc;
    rc = MPI_Bcast(w->idist->d, sizeof(mm_col_i_zone) * size, MPI_CHAR, root, comm);
    if (rc != MPI_SUCCESS) {
      fprintf(stderr, "Error Bcast w->idist->d\n");
      return NULL;
    }
    rc = MPI_Bcast(&(w->idist->N), 1, MPI_INT, root, comm);
    if (rc != MPI_SUCCESS) {
      fprintf(stderr, "Error Bcast w->idist->N\n");
      return NULL;
    }
    w->N = w->idist->N;
  }

  int myw = w->idist->d[rank].w;
  int myh = w->idist->d[rank].h;

  // working space:
  w->WA = (double*) malloc(sizeof(double) * myh * kb * kb);
  w->WB = (double*) malloc(sizeof(double) * myw * kb * kb);

  if (w->WA == NULL || w->WB == NULL) {
    fprintf(stderr, "P%d Failed to allocate matrix\n"
        "P%d myw = %d, myh = %d, kb = %d\n"
        ,rank, rank, myw, myh, kb);
    fupermod_dgemm_free_car(w);
    return NULL;
  }
  return w;
}

void fupermod_dgemm_free_car(fupermod_dgemm_workspace* w){
  free(w->WA);
  free(w->WB);
  mm_col_i_dist_free(w->idist);
  free(w); w = NULL;
}

// todo: should be someting like: 
int fupermod_dgemm_car(double *A, double *B, double *C, MPI_Comm comm, fupermod_dgemm_workspace* w) {

  int kb = w->kb;
  double* WA = w->WA;
  double* WB = w->WB;

  int rank;
  MPI_Comm_rank(comm, &rank);
  int size; 
  MPI_Comm_size(comm, &size);

  mm_col_i_distrib* idist = w->idist;
  int myx, myy, myw, myh, myi, myj;
  myx = idist->d[rank].x;
  myy = idist->d[rank].y;
  myw = idist->d[rank].w;
  myh = idist->d[rank].h;
  myi = idist->d[rank].i;
  myj = idist->d[rank].j;

  int rc;
  MPI_Comm col_comm; 
  rc = MPI_Comm_split(comm, myj, myi, &col_comm); //color: processes in 'th column, key: i'th process in column
  if (rc != MPI_SUCCESS) {
    fprintf(stderr, "Error splitting col groups\n");
    return -1;
  }	

  MPI_Comm row_comm; 
  rc = MPI_Comm_split(comm, myi, myj, &row_comm); //color: processes in 'th row, key: j'th process in row
  if (rc != MPI_SUCCESS) {
    fprintf(stderr, "Error splitting row groups\n");
    return -1;
  }

  MPI_Comm node_comm;
  int color = -1;
  int key = 0;	
  int n = 0;	
  do // colour until the current process is found
    if (n == 0 || w->idist->d[n - 1].c != w->idist->d[n].c) {
      // the next processor
      color++;
      key = 0;
    } else {
      // the same processor
      key++;
    }
  while (n++ != rank);
  rc = MPI_Comm_split(comm, color, key, &node_comm); // color: procs in the same node, key: i'th proc in column
  if (rc != MPI_SUCCESS) {
    fprintf(stderr, "Error splitting node groups\n");
    return -1;
  }

  char filename[256];
  sprintf(filename, "times.step.%d.dat", rank);
  FILE* time_step = fopen(filename, "w");
  char hostname[MPI_MAX_PROCESSOR_NAME];
  int len;
  MPI_Get_processor_name(hostname, &len);
  fprintf(time_step, "# host: %s, myw:%d myh:%d myi:%d myj:%d N:%d\n", hostname, myw, myh, myi, myj, w->N);
  fprintf(time_step, "#\tcomm\tcomp\texen\texes\tN\n");

  sprintf(filename, "times.sum.%d.dat", rank);
  FILE* time_sum = fopen(filename, "w");

  // execution time
  double sum_comm = 0;
  double sum_comp = 0;
  double sum_exen = 0;
  double sum_exes = 0;
  double* time_comm = (double *)malloc(sizeof(double) * w->N);
  double* time_comp = (double *)malloc(sizeof(double) * w->N);
  double* time_exen = (double *)malloc(sizeof(double) * w->N); // execution time of a node
  double* time_exes = (double *)malloc(sizeof(double) * w->N); // execution time of the system

  // Main k loop - moving pivrow & pivcol	
  int k;
  for (k = 0; k < w->N; k++) {
    double* A_pt = WA;
    double* B_pt = WB;
    int lda = kb, ldb = myw * kb;

    // Find j index of procs holding pivcol
    int pivcol_j = -1;
    int i;
    for (i = 0; i < size; i++) {
      if (k >= dx(i) && k < dx(i)+dw(i)) {
        pivcol_j = dj(i);
        break;
      }
    }
    // Am i pivcol?
    int pivcol = 0;
    if (myj == pivcol_j) {
      pivcol = 1;
      A_pt = A + (k - myx) * kb; // point to local A.
      lda = myw * kb;	// lda changes.
    }
    // Find i index of procs holding pivrow
    int pivrow_i = -1;
    for (i = 0; i < size; i++) {
      if (k >= dy(i) && k < dy(i)+dh(i)) {
        pivrow_i = di(i);
        break;
      }
    }
    // Am i pivrow?
    int pivrow = 0;
    if (myi == pivrow_i) {
      pivrow = 1;
      B_pt = B + (k - myy) * kb * myw * kb; // point to local B.
    }

    if (pivcol_j < 0 || pivrow_i < 0) {
      fprintf(stderr, "Error, Problem with pivcol or pivrow"
          "pivcol_j %d, pivrow_i %d\n", pivcol_j, pivrow_i);
      return -1;
    }				

    //		printf("rank %d, k %d, myi %d myj %d pivcol_j %d pivrow_i %d pivcol %d pivrow %d\n", rank, k, myi, myj, pivcol_j, pivrow_i, pivcol, pivrow);

    // move pivot column to WA
    if (pivcol) {
      for (i = 0; i < myh * kb; i++) {
        memcpy(WA + i * kb, A_pt + i * myw * kb, kb * sizeof(double));
      }			
    }
    // move pivot row to WB
    if (pivrow) {
      memcpy(WB, B_pt, kb * myw * kb * sizeof(double));
    }

    int reps = 0;
    int reps_valid = -1;
    int reps_min = 3;
    int reps_max = 55;
    int reps_warmup = 5;

    int cl = 0.95;
    int eps = 0.025;		
    int stop = 0;
    double* temp_comm = (double *)malloc(sizeof(double) * reps_max);
    double* temp_comp = (double *)malloc(sizeof(double) * reps_max);

    struct timeval t1, t2, t3, t4, t5;
    while (reps < (reps_max + reps_warmup) && (stop == 0)) {
      int res = 0; // local flag

      MPI_Barrier(comm);

      gettimeofday(&t1, NULL); // start timing

      //bcast pivot columns and rows
      rc = MPI_Bcast(WA, myh * kb * kb, MPI_DOUBLE, pivcol_j, row_comm);
      if (rc != MPI_SUCCESS) {
        fprintf(stderr, "Error Bcast B\n");
        return -1;
      }
      rc = MPI_Bcast(WB, kb * myw * kb, MPI_DOUBLE, pivrow_i, col_comm);
      if (rc != MPI_SUCCESS) {
        fprintf(stderr, "Error Bcast B\n");
        return -1;
      }

      gettimeofday(&t2, NULL); // end timing communication

      fupermod_gemm* gemm = fupermod_gemm_alloc(NULL);
      fupermod_gemm_execute(gemm, CblasRowMajor, CblasNoTrans, CblasNoTrans,  myw * kb, myh * kb, kb, 1,
          A_pt, lda, B_pt, ldb, 1, C, myw * kb);
      //dgemm('n', 'n', myw * kb, myh * kb, kb, 1, B_pt, ldb, A_pt, lda, 1, C, myw * kb);			
      fupermod_gemm_free(gemm);

      /*			if (rank == 0) {			
              printf("reps %d\n",reps);

              printf("WA(%d,%d):\n", myh * kb, kb);
              verbose_matrix(WA, myh * kb, kb);

              printf("WB(%d,%d):\n", kb, myw * kb);
              verbose_matrix(WB, kb, myw * kb);

              printf("C(%d,%d):\n", myh * kb, myw * kb);
              verbose_matrix(C, myh * kb, myw * kb);
              }
              */		
      gettimeofday(&t3, NULL); // end timing computing

      MPI_Barrier(node_comm);

      gettimeofday(&t4, NULL); // end timing the execution time of a node

      MPI_Barrier(comm);

      gettimeofday(&t5, NULL); // end timing the execution time of the system

      // statistic		
      reps_valid = reps - reps_warmup;			
      if (reps_valid >= 0) {
        double _comm = ((double)(t2.tv_sec - t1.tv_sec) + (t2.tv_usec - t1.tv_usec) * 1.0e-6);
        double _comp = ((double)(t3.tv_sec - t2.tv_sec) + (t3.tv_usec - t2.tv_usec) * 1.0e-6);
        double _exen = ((double)(t4.tv_sec - t1.tv_sec) + (t4.tv_usec - t1.tv_usec) * 1.0e-6);
        double _exes = ((double)(t5.tv_sec - t1.tv_sec) + (t5.tv_usec - t1.tv_usec) * 1.0e-6);

        time_comm[k] += temp_comm[reps_valid] = _comm;
        time_comp[k] += temp_comp[reps_valid] = _comp;
        time_exen[k] += _exen;
        time_exes[k] += _exes;				

        if (reps_valid >= (reps_min - 1)) {
          double ci_comm = fupermod_ci(cl, reps_valid, temp_comm);
          double ci_comp = fupermod_ci(cl, reps_valid, temp_comp);
          if (ci_comm * reps_valid / time_comm[k] < eps && ci_comp * reps_valid / time_comp[k] < eps) {
            res = 1;
          }
        }
        MPI_Allreduce(&res, &stop, 1, MPI_INT, MPI_MIN, comm);
      }
      reps++;			
    }
    free(temp_comm);
    free(temp_comp);

    // recording relaible timing
    time_comm[k] /= reps_valid;
    time_comp[k] /= reps_valid;
    time_exen[k] /= reps_valid;
    time_exes[k] /= reps_valid;
    fprintf(time_step, "%d\t%le\t%le\t%le\t%le\t%d\n", k, time_comm[k], time_comp[k], time_exen[k], time_exes[k], w->N);

    sum_comm += time_comm[k];
    sum_comp += time_comp[k];
    sum_exen += time_exen[k];
    sum_exes += time_exes[k];	
  }// end k loop
  fprintf(time_sum, "%le\n", sum_comm);
  fprintf(time_sum, "%le\n", sum_comp);
  fprintf(time_sum, "%le\n", sum_exen);
  fprintf(time_sum, "%le\n", sum_exes);
  fclose(time_step);
  fclose(time_sum);

  free(time_comm);
  free(time_comp);
  free(time_exen);
  free(time_exes);
  return 0;
}

void verbose_matrix(double* ptr, int height, int width) {
  int i, j;
  for (i = 0; i < height; i++) {
    for (j = 0; j < width; j++) {
      printf("  %3.3f", ptr[i * width + j]);
    }
    printf("\n");
  }
}
