//#include "config.h"
//#include "fupermod/fupermod_partition.h"
#include "mxm_car.h"
//#include "mxm_cblas.h"
#include "fupermod/fupermod_partition.h"
#include "cblas_wrappers/fupermod_cblas.h"
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define VERBOSE_MATRIX 0

/*! Fill matrix with increasing numbers - For testing purpose. */
void fill_matrix_testing(int rank, fupermod_dgemm_workspace* workspace, int kb, int N, double *A, double *B, double *C);
/*! Fill matrix with constant numbers. */
void fill_matrix_constants(int rank, fupermod_dgemm_workspace* workspace, int kb, double *A, double *B, double *C);
/*! Test: compare solution with a serial mxm. */
void compare_solution(int rank, fupermod_dgemm_workspace* workspace, int kb, int N, double *C, MPI_Comm comm);

int main(int argc, char** argv) {
  MPI_Init(&argc, &argv);
  MPI_Comm comm = MPI_COMM_WORLD;

  // exit routine
  int exit = 0; 
  // Block size
  int kb = 2;
  // gpu_id
  int gpu_id = -1;
  // Test final result.
  int test_result = 0;
  //partition file path
  char dist_path[256] = "./partition.dist";
  char conf_file[256] = "./conf_file";
  char dist_car[256] = "./partition.dist.car";

  int rank;
  MPI_Comm_rank(comm, &rank);
  int size;
  MPI_Comm_size(comm, &size);
  int root = 0;
  /*
     {	
     printf("PID %d\n", getpid());
     if (!rank)
     getc(stdin);
     MPI_Barrier(MPI_COMM_WORLD);
     }
     */	
  if (rank == root) {
    int opt;
    while ((opt = getopt(argc, argv, "ha:k:g:tp:m:")) != -1)
      switch(opt) {
        case 'h':
          fprintf(stderr,
              "routine\n"
              "Usage: [options]\n"
              "	-h	help\n"
              "	-k I	Block size (default %d)\n"
              "	-g I	Gpu id (default %d)\n"
              "	-t	Test final result (default none)\n"
              "	-p S	Path to partition file (default %s)\n"
              "	-m S	Path to machine file (default %s)\n",
              kb, gpu_id, dist_path, conf_file);
          exit = 1;
          break;
        case 'k':
          kb = atoi(optarg);
          break;
        case 'g':
          gpu_id = atoi(optarg);
          break;
        case 't':
          test_result = 1;
          break;
        case 'p':
          strcpy(dist_path, optarg);
          break;
        case 'm':
          strcpy(conf_file, optarg);
          break;
        default:
          fprintf(stderr, "Unknown option %s\n", optarg);
          break;
      }
  }

  MPI_Bcast(&exit, 1, MPI_INT, root, comm);
  if (exit) {
    MPI_Finalize();
    return 0;
  }
  MPI_Bcast(&kb, 1, MPI_INT, root, comm);
  MPI_Bcast(&test_result, 1, MPI_INT, root, comm);

  char** hostnames = fupermod_allgather_hostnames(comm);
  fupermod_process_conf* confs;
  //TODO
  fprintf(stderr, "Warning!! If you want to use this tool, implament it to read confs from machine file (see mxm routine)\n");

  fupermod_dgemm_workspace* workspace = fupermod_dgemm_alloc_car(comm, 0, kb, 0, confs, dist_car);

  if (workspace == NULL) {
    fprintf(stderr, "Error: workspace allocation\n");
    fupermod_dgemm_free_car(workspace);
    MPI_Finalize();
    return -1;	
  }
  //	if (rank == 0)
  //		mm_col_print(workspace->idist, kb);

  // Overall problem size
  long long int N = workspace->N;
  int myw = workspace->idist->d[rank].w;
  int myh = workspace->idist->d[rank].h;

  // initialise A, B, C, WA, WB
  size_t mem_size = sizeof(double) * (myw * kb * myh * kb);
  double *A, *B, *C;
  A = (double*) malloc(mem_size);
  B = (double*) malloc(mem_size);
  C = (double*) malloc(mem_size);

  // Did it work?
  if (A == NULL || B == NULL || C == NULL) {
    fprintf(stderr, "P%d Failed to allocate matrix\n"
        "P%d myw = %d, myh = %d, kb = %d\n"
        "P%d A:%p B:%p C:%p\n"
        "p%d Exiting\n",rank, rank, myw, myh, kb, rank, A, B, C, rank);
    free(A);  A = NULL; 
    free(B);  B = NULL;
    free(C);  C = NULL;

    fupermod_dgemm_free_car(workspace);
    MPI_Finalize();
    return -1;
  }
  // Fill A, B. Zero C
  if (test_result)
    fill_matrix_testing(rank, workspace, kb, N, A, B, C);
  else
    fill_matrix_constants(rank, workspace, kb, A, B, C);

  // fupermod_dgemm
  fupermod_dgemm_car(A, B, C, comm, workspace);

  if (test_result)
    compare_solution(rank, workspace, kb, N, C, comm);

  fupermod_dgemm_free_car(workspace);
  free(A);  A = NULL;
  free(B);  B = NULL;
  free(C);  C = NULL;

  MPI_Finalize();
  return 0;
}

void fill_matrix_testing(int rank, fupermod_dgemm_workspace* workspace, int kb, int N, double *A, double *B, double *C){

  int kw = kb * workspace->idist->d[rank].w;
  int kh = kb * workspace->idist->d[rank].h;
  int kx = kb * workspace->idist->d[rank].x;
  int ky = kb * workspace->idist->d[rank].y;
  int kN = kb * N;

  int i, j;
  for (i = 0; i < kh; i++) {
    for (j = 0; j < kw; j++) {
      A[i * kw + j] = (i + ky) * kN + (j + kx);
      B[i * kw + j] = (i + ky) * kN + (j + kx) + 5;
    }
  }

  for (i = 0; i < kh* kw; i++) {
    C[i] = 0.0;
  }
}

void fill_matrix_constants(int rank, fupermod_dgemm_workspace* workspace, int kb, double *A, double *B, double *C){
  int myw = workspace->idist->d[rank].w;
  int myh = workspace->idist->d[rank].h;

  int i;
  for (i = 0; i < myh * kb * myw * kb; i++) {
    A[i] = 1.1;
    B[i] = 2.5;
    C[i] = 0.0;
  }
}

void compare_solution(int rank, fupermod_dgemm_workspace* workspace, int kb, int N, double *C, MPI_Comm comm){
  int kw = kb * workspace->idist->d[rank].w;
  int kh = kb * workspace->idist->d[rank].h;
  int kx = kb * workspace->idist->d[rank].x;
  int ky = kb * workspace->idist->d[rank].y;
  int kN = kb * N;

  double* At = (double*) malloc(sizeof(double) * (kN * kN));
  double* Bt = (double*) malloc(sizeof(double) * (kN * kN));
  double* Ct = (double*) malloc(sizeof(double) * (kN * kN));

  int i, j;
  for (i = 0; i < kN * kN; i++) {
    At[i] = i;
    Bt[i] = i + 5;
  }
  for (i = 0; i < kN * kN; i++) {
    Ct[i] = 0.0;
  }

  fupermod_gemm* gemm = fupermod_gemm_alloc(NULL);
  fupermod_gemm_execute(gemm, CblasRowMajor, CblasNoTrans, CblasNoTrans, kN, kN, kN, 1,
      At, kN, Bt, kN, 0, Ct, kN);
  //dgemm('n', 'n', kN, kN, kN, 1, Bt, kN, At, kN, 0, Ct, kN); //column-major, A and B change position
  fupermod_gemm_free(gemm);

  double diff = 0;
  for (i = 0; i < kh; i++) {
    for (j = 0; j < kw; j++) {
      diff += pow(Ct[(i + ky) * kN + (j + kx)] - C[i * kw + j], 2);
    }
  }
  double sum;
  MPI_Reduce(&diff, &sum, 1, MPI_DOUBLE, MPI_SUM, 0, comm);
  if (rank == 0) {
    if (sum < 1E-6)
      fprintf(stderr, "Matrix multiplication was successful\n");
    else 
      fprintf(stderr, "***\nError: Matrix multiplication failed, diff = %le\n***\n", sum);
  }

  free(At); At = NULL;
  free(Bt); Bt = NULL;
  free(Ct); Ct = NULL;
}

