/*!
 * \page generate_conf Genearate Conf
 This tool generates a default conf file for the processes in an mpirun execution.
 */

#include "config.h"
#include "fupermod/fupermod_conf.h"
#include <string.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>

int main(int argc, char** argv) {

  char subopts[256]   = "";
  char* conf_file = "./conf_file";
  MPI_Init(&argc, &argv);
  MPI_Comm comm = MPI_COMM_WORLD;
  int exit = 0;


  int rank;
  MPI_Comm_rank(comm, &rank);
  int root = 0;
  int opt;
  while ((opt = getopt(argc, argv, "hl:f:o:m:L:U:g:c:s:r:R:T:i:e:z:Z:d")) != -1){
    switch (opt) {
      case 'h':
        if (rank == root) {
          fprintf(
              stderr,
              "Generates a default conf file\n"
              "Usage: generate_conf [options]\n"
              "	-h	help\n"
              "	-o S	suboptions for the shared library separated by comma\n");
        }
        exit = 1;
        break;
      case 'o':
        strcpy(subopts, optarg);
        break;
      default:
        fprintf(stderr, "Unknown option %s\n", optarg);
        break;
    }
  }
  MPI_Bcast(&exit, 1, MPI_INT, root, comm);
  if (exit) {
    MPI_Finalize();
    return 0;
  }

  FILE* file = fopen(conf_file, "w");
  MPI_Barrier(MPI_COMM_WORLD);
  if(file == NULL) {
    perror("can't create conf_file");
    MPI_Finalize();
    return 0;
  }
  fupermod_print_conf(MPI_COMM_WORLD, root, file, "cpu", subopts);
  fclose(file);

  MPI_Finalize();
  return 0;
}
