/*!
 * \page model_smoother Model smoother
 */
#include "config.h"
#include "fupermod/fupermod_model.h"
#include <getopt.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

int main(int argc, char** argv) {
  MPI_Init(&argc, &argv);
  MPI_Comm comm = MPI_COMM_SELF;

  char path[256] = "", subopts[256] = "";
  char data_dir_in[256] = "./fpm";
  char data_dir_out[256] = "./fpm.smooth";
  char conf_file[256] = "conf_file";
  int avgc = 10;

  int opt;
  while ((opt = getopt(argc, argv, "hl:o:f:i:m:c:")) != -1)
    switch (opt) {
      case 'h':
        fprintf(
            stderr,
            "Partitions data, building functional performance models\n"
            "Usage: partitioner [options]\n"
            "-h	help\n"
            "-l S	shared library (required: a full path or relative to LD_LIBRARY_PATH)\n"
            "-o S	suboptions (comma separated options for the shared library)\n"
            "-m S path to conf_file, specify machine fpm's to use.\n" 
            "-i S path to input data file (default: %s)\n"
            "-f S path to output data file (default: %s)\n"
            "-c I number of points to smooth over (default: %d)\n"
            , data_dir_in, data_dir_out, avgc
            );
        return 0;
        break;
      case 'l':
        strcpy(path, optarg);
        break;
      case 'o':
        strcpy(subopts, optarg);
        break;
      case 'm':
        strcpy(conf_file, optarg);
        break;
      case 'i':
        strcpy(data_dir_in, optarg);
        break;
      case 'f':
        strcpy(data_dir_out, optarg);
        break;
      case 'c':
        avgc = atoi(optarg);
        break;
      default:
        fprintf(stderr, "Unknown option %s\n", optarg);
        break;
    }
  if (!strcmp(path, "")) {
    fprintf(stderr, "Routine shared library must be specified\n");
    return -1;
  }

  fupermod_library library;
  fupermod_library_load(comm, path, NULL, &library);

  int size = 0;
  // get the host names
  if (!strcmp(conf_file, "")) {
    fprintf(stderr, "Error: must specify conf_file\n");
    return -1;
  }
  fupermod_process_conf conf = fupermod_get_conf(MPI_COMM_WORLD, conf_file);


  int p;
  for (p = 0; p < size; p++) {
    char* filename = fupermod_data_filename(data_dir_in, conf);
    FILE* data_file = fopen(filename, "r");
    if (data_file == NULL){
      fprintf(stderr,"Error: could not open input: %s\n", filename);
      return -1;
    }
    free(filename); filename = NULL;
    fupermod_data* data_in;	
    data_in = fupermod_data_alloc();
    fupermod_data_read(data_file, data_in);
    fclose(data_file);

    //todo: change hostname to host_device_name
    filename = fupermod_data_filename(data_dir_out, conf);
    FILE* outfile = fopen(filename, "w");
    if (outfile == NULL){
      fprintf(stderr,"Error: could not open output: %s\n", filename);
      return -1;
    }
    free(filename); filename = NULL;

    fupermod_data_write_head(outfile, &library);

    int i;
    for (i = 0; i < data_in->count; i++){
      fupermod_point point;
      double sum = 0;
      int avg = 0;
      int j;
      for (j = -avgc / 2; j <= avgc / 2; j++){
        if ((i + j) >= 0 && (i + j) < data_in->count) {
          sum += library.kernel->complexity(data_in->points[i + j].d, NULL) / data_in->points[i + j].t;
          avg++;
          /*
             } else {
             sum += data_in->points[i].s;
             avg++;
             */
        }
      }
      double s_avg = sum / avg;
      point.d = data_in->points[i].d;
      point.t = library.kernel->complexity(point.d, NULL) / s_avg;
      fupermod_point_write(outfile, &point, library.kernel->complexity);
    }
    fupermod_data_free(data_in);
    fclose(outfile);
  }

  fupermod_library_unload(comm, &library);

  MPI_Finalize();
  return 0;
}
