/*!
 * \page modeller Model test
 * Test for approximations of time functions
 */
#include "config.h"
#include "fupermod/fupermod.h"
#include <getopt.h>
#include <stdio.h>
#include <string.h>
#include <dlfcn.h>
#include <stdlib.h>

int main(int argc, char** argv) {
  char path[256] = "";
  char subopts[256] = "";
  int M = 20;
  int N = 20;
  int x = 10;
  int y = 10;

  char data_file[256] = "";
  char name[MPI_MAX_PROCESSOR_NAME] = "";
  MPI_Comm comm = MPI_COMM_SELF;

  int opt;
  while ((opt = getopt(argc, argv, "hl:o:M:N:x:y:f:")) != -1)
    switch (opt) {
      case 'h':
        fprintf(
            stderr,
            "Fit models to data points\n"
            "Usage: partitioner [options]\n"
            "-h		help\n"
            "-l S	shared library (required: a full path or relative to LD_LIBRARY_PATH)\n"
            "-o S	suboptions (comma separated options for the shared library)\n"
            "-M I	problem size to model (default %d)\n"
            "-N I	problem size to model (default %d)\n"
            "-x I	grid size in M direction (default %d)\n"
            "-y I	grid size in N direction (default %d)\n"
            "-f S	path to data files\n"
            , M, N, x, y
            );
        return 0;
        break;
      case 'l':
        strcpy(path, optarg);
        break;
      case 'o':
        strcpy(subopts, optarg);
        break;
      case 'M':
        M = atoi(optarg);
        break;
      case 'N':
        N = atoi(optarg);
        break;
      case 'x':
        x = atoi(optarg);
        break;
      case 'y':
        y = atoi(optarg);
        break;
      case 'f':
        strcpy(data_file, optarg);
        break;
      default:
        fprintf(stderr, "Unknown option %s\n", optarg);
        break;
    }

  if (!strcmp(data_file, "")) {
    fprintf(stderr, "data_file not specified\n");
    return -1;
  }
  if (!strcmp(path, "")) {
    fprintf(stderr, "%s: routine shared library not specified\n", name);
    return -1;
  }
  fupermod_library library;
  fupermod_library_load(comm, path, NULL, &library);
  //fupermod_kernel* kernel = library.loader->kernel;

  //fupermod_2d_data* data = fupermod_2d_data_alloc();
  //FILE* data_file = fopen(data_file, "r");

  //if (file == NULL) {
  //	fprintf(stderr, "Error reading data file\n");
  //	return -1;
  //}
  //fupermod_2d_data_read(data_file, data);
  //fclose(data_file);
  ////	fupermod_model* model = fupermod_model_fit_alloc(data, N, &((fupermod_routine*)kernel)->kernel, nbreak);
  //fupermod_model* model = fupermod_model_interp_alloc(data, N, &((fupermod_routine*)kernel)->kernel, smooth);

  //FILE* file = fopen("model_fit.dat", "w");
  //fprintf(file, "#kernel:%s, N:%d, machine:%s\n", kernel->name, N, name);
  //fupermod_print_models(file, 1, &model, N); 
  //fclose(file);

  ////fupermod_model_fit_free(model);
  //fupermod_model_interp_free(model);
  //fupermod_data_free(data);
  //err = kernel->unload(comm);
  //if (err) {
  //	fprintf(stderr, "%s: kernel finalize failed\n", name);
  //	return -1;
  //}

  fupermod_library_unload(comm, &library);

  return 0;
}
